/*
This function loads a bitmap for a desktop icon
Arguments: pointer to the pointer where to allocate memory,
string that contains a path to the icon file
*/

unsigned load_icon_bitmap(unsigned char **icon, char *icon_file)
{
  FILE *fp;
  *icon = malloc(100*sizeof(char));

  if(!icon) return 0;
  
  if(fp=fopen(icon_file, "rb"))
  { 
    fread(*icon, sizeof(char), 100, fp);
    fclose(fp);
  }
  else
  {
    free(icon);
    return 0;
  }
  return 1;
}

/*
Loads icon
arguments: pointer to the icon struct, path to the icon bitmap file,
a string that contains the text of icon
*/
void load_icon(struct icon *icon, char *icon_file, unsigned char *text)
{
  load_icon_bitmap(&icon->bitmap, icon_file);
  if(text)
  {  
    icon->text = malloc((1+strlen(text))*sizeof(char));
//    while(*tp++ = *text++);
    strcpy(icon->text, text);
  }
}
 
void init_desktop(void)
{
  char *static_icons[][2] =
  {
    {
      "icons/runprmpt.ico", "Run program"
    },
    {
      "icons/fileman.ico", "File manager"
    },
    {
      "icons/cpanel.ico", "Control panel"
    },
    {
      "icons/clock.ico", "Clock"
    },
    {
      "icons/vpucon.ico", "VPU Console"
    }
  };
  unsigned x = 20;
//  windows = malloc(sizeof(struct prog_window*));
//  window_visibility_order = malloc(sizeof(int));
//  desktop_icons = malloc((sizeof(static_icons)/sizeof(char*))*sizeof(struct icon));
  desktop_icons = malloc(((sizeof(static_icons)>>1)/sizeof(char*))*sizeof(struct icon));
  default_icons = malloc(sizeof(char**));

  default_icons[0] = malloc(sizeof(char*));
  load_icon_bitmap(&default_icons[0], "icons/wndwdflt.ico");

  // load static icons
  for(;desktop_icon_count<(sizeof(static_icons)>>1)/sizeof(char*);desktop_icon_count++,x+=60)
  {
    load_icon(&desktop_icons[desktop_icon_count], static_icons[desktop_icon_count][0], static_icons[desktop_icon_count][1]);
    desktop_icons[desktop_icon_count].x=x;
    desktop_icons[desktop_icon_count].y=20;    
  }  
}

#define _ICON_SHADOW_COLOR _WINDOW_BGCOLOR

void draw_icon(unsigned int x, unsigned int y, unsigned char *icon, unsigned char *text, unsigned char color)
{
  unsigned int n=20;
  unsigned int m;
  unsigned int xc;
  unsigned shadowcolor;

  if(color == _DESKTOP_BGCOLOR)
    shadowcolor = _DESKTOP_BGCOLOR;
  else
    shadowcolor = _ICON_SHADOW_COLOR;

  while(n--)
  {
    for(m=5,xc=x;m--;)
    {
      if(*icon & 0x80)
        put_pixel_shadow(xc, y, color, shadowcolor);
      xc++;
      if(*icon & 0x40)
        put_pixel_shadow(xc, y, color, shadowcolor);
      xc++;      
      if(*icon & 0x20)
        put_pixel_shadow(xc, y, color, shadowcolor);
      xc++;
      if(*icon & 0x10)
        put_pixel_shadow(xc, y, color, shadowcolor);
      xc++;
      if(*icon & 0x08)
        put_pixel_shadow(xc, y, color, shadowcolor);
      xc++;
      if(*icon & 0x04)
        put_pixel_shadow(xc, y, color, shadowcolor);
      xc++;
      if(*icon & 0x02)
        put_pixel_shadow(xc, y, color, shadowcolor);
      xc++;
      if(*icon & 0x01)
        put_pixel_shadow(xc, y, color, shadowcolor);
      xc++;
      icon++;
    }
    y++;
  }

  gprint_text(x+1, y+17-_FONT_HEIGHT, text, shadowcolor, _FONT_HEIGHT, x+_WINDOW_ICONS_MARGIN, 0);  
  gprint_text(x, y+16-_FONT_HEIGHT, text, color, _FONT_HEIGHT, x+_WINDOW_ICONS_MARGIN, 0);
}

void save_desktop_shortcut(unsigned shortcut)
{
  char coords[5];
  unsigned n;
  FILE *fp;
  char shortcut_file_path[81];

  sprintf(shortcut_file_path, "DESKTOP\\%s", desktop_shortcuts[shortcut].shortcut_file);  
  
  fp = fopen(shortcut_file_path, "w");
  
  for(n=0;desktop_shortcuts[shortcut].path[n];n++)
    fputc(desktop_shortcuts[shortcut].path[n], fp);
  fputc(';', fp);
  
  for(n=0;desktop_shortcuts[shortcut].text[n];n++)
    fputc(desktop_shortcuts[shortcut].text[n], fp);
  fputc(';', fp);

  sprintf(coords, "%i", desktop_shortcuts[shortcut].x);
  for(n=0;coords[n];n++)
    fputc(coords[n], fp);
  fputc(';', fp);
  
  sprintf(coords, "%i", desktop_shortcuts[shortcut].y);
  for(n=0;coords[n];n++)
    fputc(coords[n], fp);
  fputc(';', fp);
  
  sprintf(coords, "%i", desktop_shortcuts[shortcut].ask_args);
  for(n=0;coords[n];n++)
    fputc(coords[n], fp);
  fputc(';', fp);
  
  if(desktop_shortcuts[shortcut].icon_path)
  {
    for(n=0;desktop_shortcuts[shortcut].icon_path[n];n++)
      fputc(desktop_shortcuts[shortcut].icon_path[n], fp);
  }

  fclose(fp);
}

unsigned load_desktop_shortcuts(void)
{
  FILE *fp;
  FILE *icon;
  unsigned int m;
  int c;
  char coords[5];
  char icon_path[81];
  struct find_t fileinfo;
  unsigned rc;
  if(!(desktop_shortcuts = malloc(sizeof(struct desktop_shortcut))))
    return 0;

  rc = _dos_findfirst("DESKTOP\\*.lnk", _A_NORMAL, &fileinfo); 
  while(!rc)
  {
    if(shortcut_count)
    {
      if(!(desktop_shortcuts = realloc(desktop_shortcuts, (1+shortcut_count)*sizeof(struct desktop_shortcut))))
        return 0;
    }
      
    sprintf(icon_path, "DESKTOP\\%s", fileinfo.name);
    fp = fopen(icon_path, "r");

    sprintf(desktop_shortcuts[shortcut_count].shortcut_file, "%s", fileinfo.name);

    for(m=0;(c = fgetc(fp)) != ';';m++)
      desktop_shortcuts[shortcut_count].path[m] = c;
    desktop_shortcuts[shortcut_count].path[m]=0;

    for(m=0;(c = fgetc(fp)) != ';';m++)
      desktop_shortcuts[shortcut_count].text[m] = c;
    desktop_shortcuts[shortcut_count].text[m]=0;      

    for(m=0;(c = fgetc(fp)) != ';';m++)
      coords[m] = c;
    coords[m] = 0;
    desktop_shortcuts[shortcut_count].x = atoi(coords);

    for(m=0;(c = fgetc(fp)) != ';';m++)
      coords[m] = c;
    coords[m] = 0;      
    desktop_shortcuts[shortcut_count].y = atoi(coords);

    for(m=0;(c = fgetc(fp)) != ';';m++)
      coords[m] = c;
    coords[m] = 0;      
    desktop_shortcuts[shortcut_count].ask_args = atoi(coords);    

    for(m=0;(c = fgetc(fp)) != -1;m++)
      icon_path[m] = c;
    icon_path[m] = 0;

    fclose(fp);

    if(icon = fopen(icon_path, "rb"))
    {
      if(!(desktop_shortcuts[shortcut_count].icon = malloc(100*sizeof(char))))
        return 0;
      fread(desktop_shortcuts[shortcut_count].icon, sizeof(char), 100, icon);
      fclose(icon);
      
      if(!(desktop_shortcuts[shortcut_count].icon_path = malloc(81*sizeof(char))))
        return 0;
      sprintf(desktop_shortcuts[shortcut_count].icon_path, "%s", icon_path);
    }
    else
    {
      desktop_shortcuts[shortcut_count].icon = 0;
      desktop_shortcuts[shortcut_count].icon_path = 0;
    }

    shortcut_count++;
    rc = _dos_findnext(&fileinfo);     
  }
  return 1;
}

/*void draw_desktop_shortcuts(void)
{
  unsigned int n = shortcut_count;
  unsigned char color;
  while(n--)
  {
    if(icon_selected == desktop_icon_count + n)
    {
      color = _ICON_ACOLOR;
      draw_borders(desktop_shortcuts[n].x, desktop_shortcuts[n].y, 40, 35, 2);            
    }
    else
    {
      color = _ICON_ICOLOR;
      draw_borders(desktop_shortcuts[n].x, desktop_shortcuts[n].y, 40, 35, 4);                  
    }
      
    if(desktop_shortcuts[n].icon)
      draw_icon(desktop_shortcuts[n].x, desktop_shortcuts[n].y, desktop_shortcuts[n].icon, desktop_shortcuts[n].text, color);
    else
      draw_icon(desktop_shortcuts[n].x, desktop_shortcuts[n].y, default_icons[0], desktop_shortcuts[n].text, color);    
  }
}*/

void draw_desktop_icon(unsigned n, unsigned undraw)
{
  unsigned x;
  unsigned y;
  unsigned char *bitmap;
  char *text;
  unsigned char color;
  void *ptr = running_window;


  if(n >= desktop_icon_count)
  {
    x = desktop_shortcuts[n-desktop_icon_count].x;
    y = desktop_shortcuts[n-desktop_icon_count].y;
    text = desktop_shortcuts[n-desktop_icon_count].text;
    if(desktop_shortcuts[n-desktop_icon_count].icon)
      bitmap = desktop_shortcuts[n-desktop_icon_count].icon;
    else
      bitmap = default_icons[0];
  }
  else
  {
    x = desktop_icons[n].x;
    y = desktop_icons[n].y;
    bitmap = desktop_icons[n].bitmap;
    text = desktop_icons[n].text;
  }
  running_window = (void*)-1;
  if(undraw)
  {
    color = _DESKTOP_BGCOLOR;
    draw_borders(x, y, 40, 35, 4);                            
  } 
  else if(icon_selected == n)
  {
    color = _ICON_ACOLOR;
    draw_borders(x, y, 40, 35, 2);                  
  }
  else
  {
    color = _ICON_ICOLOR;
    draw_borders(x, y, 40, 35, 4);                        
  }
  draw_icon(x, y, bitmap, text, color);
  running_window = ptr;
}

void move_icon(unsigned icon, int dx, int dy)
{
  unsigned oldx, oldy;

  draw_desktop_icon(icon, 1);  
  if(icon < desktop_icon_count)
  {
    oldx = desktop_icons[icon].x;
    oldy = desktop_icons[icon].y;
    desktop_icons[icon].x += dx;
    desktop_icons[icon].y += dy;
  }
  else
  {
    oldx = desktop_shortcuts[icon-desktop_icon_count].x;
    oldy = desktop_shortcuts[icon-desktop_icon_count].y;
    desktop_shortcuts[icon-desktop_icon_count].x += dx;
    desktop_shortcuts[icon-desktop_icon_count].y += dy;
    if(desktop_shortcuts[icon-desktop_icon_count].x < 0 || desktop_shortcuts[icon-desktop_icon_count].x > _RES_X - 30)
      desktop_shortcuts[icon-desktop_icon_count].x = _RES_X - 30;
    if(desktop_shortcuts[icon-desktop_icon_count].y < 0 || desktop_shortcuts[icon-desktop_icon_count].y > _RES_Y - 30)
      desktop_shortcuts[icon-desktop_icon_count].y = _RES_Y - 30;      
  }
  draw_desktop_icon(icon, 0);
}

void draw_select_icon(unsigned old, unsigned new)
{
  draw_desktop_icon(old, 0);
  draw_desktop_icon(new, 0);
}

void draw_desktop_icons(void)
{
  unsigned int n = desktop_icon_count + shortcut_count;

  while(n--)
    draw_desktop_icon(n, 0);
}

void draw_window_icons(void)
{
  unsigned int n;
  unsigned int x = _WINDOW_ICONS_LEFT;
  for(n=0;n<window_count;n++)
  {
    if(windows[n]->minimized)
    {
      draw_borders(x, _RES_Y-40, _WINDOW_ICONS_MARGIN-2, 39, 0);
      if(windows[n]->icon)
        draw_icon(x, _RES_Y-40, windows[n]->icon, windows[n]->title, _WINDOW_ICON_COLOR);
      else
        draw_icon(x, _RES_Y-40, default_icons[0], windows[n]->title, _WINDOW_ICON_COLOR);
      x+=_WINDOW_ICONS_MARGIN;
    }
  }  
}

void draw_desktop(void)
{
  unsigned int x=0;
  unsigned int y=0;

/*  if(_VIDEO_MODE == 0x10 || _VIDEO_MODE == 0x12)
    fill_screen_ega4(_DESKTOP_BGCOLOR);
  else if(_VIDEO_MODE == 0x0F || _VIDEO_MODE == 0x11)
    fill_screen_ega2(_DESKTOP_BGCOLOR);
  else if(_VIDEO_MODE == 0x06)
    fill_screen_cga(_DESKTOP_BGCOLOR);
  else
    for(;y<_RES_Y;y++,x=0)
      for(;x<_RES_X;x++)
        put_pixel(x, y, _DESKTOP_BGCOLOR);*/
  if(!wallpaper_struct.bitmap)
//    fill_screen_color(_DESKTOP_BGCOLOR);
    call_fill_screen(_DESKTOP_BGCOLOR, FP_SEG(&fill_screen_color));
  else
    draw_xpm_bitmap(&wallpaper_struct, 0, 0, _RES_X, _RES_Y);

  draw_desktop_icons();
//  draw_desktop_shortcuts();
}

void draw_windows(void)
{
  unsigned int n;
  draw_window_icons();  
  for(n=0;n<window_count;n++)
  {
    if(!windows[window_visibility_order[n]]->minimized && !windows[window_visibility_order[n]]->hidden)            
      if(active_window == window_visibility_order[n])
        draw_window(windows[window_visibility_order[n]], 1, 0);
      else
        draw_window(windows[window_visibility_order[n]], 0, 0);
  }
}

void draw_screen(void)
{
  if(_VIDEO_MODE != video)
    restore_videomode();
  draw_desktop();
  draw_windows();
}

void update_titlebar(struct prog_window *window)
{
  if(titlebar_drag)
  {
    titlebar_drag->x0 = window->x;
    titlebar_drag->y0 = window->y;
    titlebar_drag->x1 = window->x+window->width;
    titlebar_drag->y1 = window->y+_FONT_HEIGHT+2;
  }
}

void set_active_window(int window, unsigned draw)
{
  unsigned int n = window;
  if(window>=0)
  {
    while(windows[window]->hidden)
    {
      window--;
      if(window == -1)
        window = window_count-1;
      if(window == n)
      {
        active_window = -1;
        return;
      }
    }
    
    for(n=0;n<window_count;n++)
      if(window_visibility_order[n] == window)
        break;
    for(;n<window_count-1;n++)
      window_visibility_order[n]=window_visibility_order[n+1];
    window_visibility_order[n]=active_window=window;

    update_titlebar(windows[window]);
    
    if(window_count>1 && !windows[window_visibility_order[n-1]]->minimized
    && !windows[window_visibility_order[n-1]]->hidden && draw)
      draw_window(windows[window_visibility_order[n-1]], 0, 2);
    if(window_visibility_order[n-1] != active_window && draw)
      draw_window(windows[active_window], 1, 0);
  }
  else
    active_window=window;
}
