void vpu_set_sign(struct vpu *vpu, unsigned char reg0)
{
  if((!(reg0 & 0x80) && vpu->byteregs[reg0 & 0x0F] & 0x80) || ((reg0 & 0x80) && vpu->regs[reg0 & 0x07] & 0x8000))
    vpu->flags.sign = 1;
  else vpu->flags.sign = 0; 
}

unsigned is_same_sign(struct vpu *vpu, unsigned char reg0, unsigned char reg1)
{
  if((reg0 & 0x80) && (reg1 & 0x80))
    if((vpu->regs[reg0 & 0x07] & 0x8000) == (vpu->regs[reg1 & 0x07] & 0x8000))
      return 1;
  if(!(reg0 & 0x80) && !(reg1 & 0x80))
    if((vpu->byteregs[reg0 & 0x0F] & 0x80) == (vpu->byteregs[reg1 & 0x0F] & 0x80))
      return 1;
  if((reg0 & 0x80) && !(reg1 & 0x80))
    if(((vpu->regs[reg0 & 0x07] & 0x8000) >> 8) == (vpu->byteregs[reg1 & 0x0F] & 0x80))
      return 1;
  if(!(reg0 & 0x80) && (reg1 & 0x80))
    if((vpu->byteregs[reg0 & 0x0F] & 0x80) == ((vpu->regs[reg1 & 0x07] & 0x8000) >> 8))
      return 1;
  return 0;
}

void vpu_instr_add(struct vpu *vpu)
{
  unsigned char reg0 = vpu->code[vpu->code_segment][vpu->ip++];
  unsigned char reg1 = vpu->code[vpu->code_segment][vpu->ip++];
  unsigned temp;
  unsigned char prep_overflow = is_same_sign(vpu, reg0, reg1);
  
  if(reg0 & 0x80 && reg1 & 0x80)
  {
    vpu->regs[reg0 & 0x07] += vpu->regs[reg1 & 0x07];
    if(vpu->regs[reg0 & 0x07] < vpu->regs[reg1 & 0x07])
      vpu->flags.carry=1;
  }

  else if(reg0 & 0x80)
  {
    vpu->regs[reg0 & 0x07] += vpu->byteregs[reg1 & 0x0F];
    if(vpu->regs[reg0 & 0x07] < vpu->byteregs[reg1 & 0x0F])
      vpu->flags.carry=1;
  }

  else if(reg1 & 0x80)
  {
    if((temp = vpu->byteregs[reg0 & 0x0F] + vpu->regs[reg1 & 0x07]) > 0xFF)
      vpu->flags.carry=1;
    vpu->byteregs[reg0 & 0x0F] = temp;
  }

  else
  {
    vpu->byteregs[reg0 & 0x0F] += vpu->byteregs[reg1 & 0x0F];
    if(vpu->byteregs[reg0 & 0x0F] < vpu->byteregs[reg1 & 0x0F])
      vpu->flags.carry=1;
  }
  
  if(prep_overflow != is_same_sign(vpu, reg0, reg1))
    vpu->flags.overflow=1;
  vpu_set_sign(vpu, reg0);
}

void vpu_instr_sub(struct vpu *vpu)
{
  unsigned char reg0 = vpu->code[vpu->code_segment][vpu->ip++];
  unsigned char reg1 = vpu->code[vpu->code_segment][vpu->ip++];
  unsigned char prep_overflow = is_same_sign(vpu, reg0, reg1);
  
  if(reg0 & 0x80 && reg1 & 0x80)
  {
    if(vpu->regs[reg1 & 0x07] > vpu->regs[reg0 & 0x07])
      vpu->flags.carry = 1;
    vpu->regs[reg0 & 0x07] -= vpu->regs[reg1 & 0x07];
  }

  else if(reg0 & 0x80)
  {
    if(vpu->byteregs[reg1 & 0x0F] > vpu->regs[reg0 & 0x07])
      vpu->flags.carry = 1;
    vpu->regs[reg0 & 0x07] -= vpu->byteregs[reg1 & 0x0F];
  }

  else if(reg1 & 0x80)
  {
    if(vpu->regs[reg1 & 0x07] > vpu->byteregs[reg0 & 0x0F])
      vpu->flags.carry = 1;
    vpu->byteregs[reg0 & 0x0F] -= vpu->regs[reg1 & 0x07];
  }

  else
  {
    if(vpu->byteregs[reg1 & 0x0F] > vpu->byteregs[reg0 & 0x0F])
      vpu->flags.carry = 1;
    vpu->byteregs[reg0 & 0x0F] -= vpu->byteregs[reg1 & 0x0F];
  }

  if(prep_overflow != is_same_sign(vpu, reg0, reg1))
    vpu->flags.overflow=1;
  vpu_set_sign(vpu, reg0);
}
