void vpu_instr_call(struct vpu *vpu)
{
  unsigned short new_segment;

  if(debug)
  {
    sprintf(message, "  CALL FROM segment 0x%.4X address 0x%.4X -> ", vpu->code_segment, vpu->ip);
    putstr(message);
  }

  if(vpu->code[vpu->code_segment][vpu->ip] & 0x80)
  {
    new_segment = *(unsigned short*)&vpu->code[vpu->code_segment][vpu->ip] & 0x7F;
    if(new_segment > vpu->code_count)
    {
//      sprintf(message, "Program tried to CALL JUMP to unallocated segment!\n");
//      putstr(message);
      putstr(segfaulterror);
      send_vpu_signal(vpu, SIGSEGV);    
      return;
    }
    vpu->ip += 2;
    if(debug)
    {
      sprintf(message, "segment 0x%.4X ", new_segment);
      putstr(message);
    }
  }
  else
    vpu->ip++;

  if(vpu->sp <= 2)
  {
//    sprintf(message, "Stack overflow!\n");
//    putstr(message);
    putstr(stackoverflowerror);
    send_vpu_signal(vpu, SIGSEGV);        
    return;
  }
  vpu->stack[vpu->sp--] = vpu->ip+2;
  vpu->stack[vpu->sp--] = vpu->code_segment;

  vpu->ip = *(unsigned short*)&vpu->code[vpu->code_segment][vpu->ip];
  vpu->code_segment = new_segment;
  if(debug)
  {
    sprintf(message, "offset 0x%.4X", vpu->ip);
    putstr(message);
  }
}

void vpu_instr_ret(struct vpu *vpu)
{
  if(vpu->stack_size - vpu->sp < 3)
  {
//    sprintf(message, "Stack underflow!\n");
//    putstr(message);
    putstr(stackunderflowerror);
    send_vpu_signal(vpu, SIGSEGV);        
    return;
  }
  vpu->code_segment = vpu->stack[++vpu->sp];
  vpu->ip = vpu->stack[++vpu->sp];

  if(debug)
  {
    sprintf(message, "  RET to segment 0x%.4X offset 0x%.4X", vpu->code_segment, vpu->ip);
    putstr(message);
  }

  vpu->wait = vpu->retwait;
}
