void vpu_instr_mul(struct vpu *vpu)
{
  unsigned char reg0 = vpu->code[vpu->code_segment][vpu->ip++];
  unsigned char reg1 = vpu->code[vpu->code_segment][vpu->ip++];
  uint32_t temp;
  unsigned char prep_overflow = is_same_sign(vpu, reg0, reg1);

  if(debug)
  {
    sprintf(message, "  Instruction: MUL 0x%X 0x%X\n", reg0, reg1);
    putstr(message);
  }

  if(reg0 & 0x80 && reg1 & 0x80)
  {
//    vpu->regs[reg0 & 0x0F] *= vpu->regs[reg1 & 0x0F];
    temp = vpu->regs[reg0 & 0x07] * vpu->regs[reg1 & 0x07];
    vpu->regs[reg0 & 0x07] = temp;
    if(vpu->regs[reg0 & 0x07] != temp)
      vpu->flags.carry = 1;
    return;
  }

  else if(reg0 & 0x80)
  {
    temp = vpu->regs[reg0 & 0x07] * vpu->byteregs[reg1 & 0x0F];
    vpu->regs[reg0 & 0x07] = temp;
    if(vpu->regs[reg0 & 0x07] != temp)
      vpu->flags.carry = 1;
    return;
  }

  else if(reg1 & 0x80)
    temp = vpu->byteregs[reg0 & 0x0F] * vpu->regs[reg1 & 0x07];

  else
    temp = vpu->byteregs[reg0 & 0x0F] * vpu->byteregs[reg1 & 0x0F];

  vpu->byteregs[reg0 & 0x0F] = temp;
  if(vpu->byteregs[reg0 & 0x0F] != temp)
    vpu->flags.carry = 1;

  if(prep_overflow != is_same_sign(vpu, reg0, reg1))
    vpu->flags.overflow=1;
  vpu_set_sign(vpu, reg0);  
}

void vpu_instr_div(struct vpu *vpu)
{
  unsigned char reg0 = vpu->code[vpu->code_segment][vpu->ip++];
  unsigned char reg1 = vpu->code[vpu->code_segment][vpu->ip++];
  unsigned short operand0, operand1;

  if(debug)
  {
    sprintf(message, "  Instruction: DIV 0x%X 0x%X\n", reg0, reg1);
    putstr(message);
  }

  if(reg0 & 0x80 && reg1 & 0x80)
  {
    operand0 = vpu->regs[reg0 & 0x07];
    operand1 = vpu->regs[reg1 & 0x07];
    if(!operand1)
    {
//      vpu->flags.divzero = 1;
      send_vpu_signal(vpu, SIGFPE);
      return;
    }
    vpu->regs[reg0 & 0x07] = operand0 / operand1;
    vpu->regs[reg1 & 0x07] = operand0 % operand1;
  }

  else if(reg0 & 0x80)
  {
    operand0 = vpu->regs[reg0 & 0x07];
    operand1 = vpu->byteregs[reg1 & 0x0F];
    if(!operand1)
    {
//      vpu->flags.divzero = 1;
      send_vpu_signal(vpu, SIGFPE);
      return;
    }
    vpu->regs[reg0 & 0x07] = operand0 / operand1;
    vpu->byteregs[reg1 & 0x0F] = operand0 % operand1;
  }

  else if(reg1 & 0x80)
  {
    operand0 = vpu->byteregs[reg0 & 0x0F];
    operand1 = vpu->regs[reg1 & 0x07];
    if(!operand1)
    {
//      vpu->flags.divzero = 1;
      send_vpu_signal(vpu, SIGFPE);
      return;
    }
    vpu->byteregs[reg0 & 0x0F] = operand0 / operand1;
    vpu->regs[reg1 & 0x07] = operand0 % operand1;
  }

  else
  {
    operand0 = vpu->byteregs[reg0 & 0x0F];
    operand1 = vpu->byteregs[reg1 & 0x0F];
    if(!operand1)
    {
//      vpu->flags.divzero = 1;
      send_vpu_signal(vpu, SIGFPE);
      return;
    }
    vpu->byteregs[reg0 & 0x0F] = operand0 / operand1;
    vpu->byteregs[reg1 & 0x0F] = operand0 % operand1;
  }

  vpu_set_sign(vpu, reg0);
}

void vpu_instr_idiv(struct vpu *vpu)
{
  unsigned char reg0 = vpu->code[vpu->code_segment][vpu->ip++];
  unsigned char reg1 = vpu->code[vpu->code_segment][vpu->ip++];
  signed short operand0, operand1;
  signed short *reswp;
  signed short *remwp;
  signed char *resbp;
  signed char *rembp;
  unsigned char prep_overflow = is_same_sign(vpu, reg0, reg1);  

  if(debug)
  {
    sprintf(message, "  Instruction: IDIV 0x%X 0x%X\n", reg0, reg1);
    putstr(message);
  }

  if(reg0 & 0x80 && reg1 & 0x80)
  {
    operand0 = *(signed short*)&vpu->regs[reg0 & 0x07];
    operand1 = *(signed short*)&vpu->regs[reg1 & 0x07];
    if(!operand1)
    {
//      vpu->flags.divzero = 1;
      send_vpu_signal(vpu, SIGFPE);
      return;
    }
    reswp = (signed short*)&vpu->regs[reg0 & 0x07];
    remwp = (signed short*)&vpu->regs[reg1 & 0x07];
    *reswp = operand0 / operand1;
    *remwp = operand0 % operand1;
  }

  else if(reg0 & 0x80)
  {
    operand0 = *(signed short*)&vpu->regs[reg0 & 0x07];
    operand1 = *(signed char*)&vpu->byteregs[reg1 & 0x0F];
    if(!operand1)
    {
//      vpu->flags.divzero = 1;
      send_vpu_signal(vpu, SIGFPE);
      return;
    }
    reswp = (signed short*)&vpu->regs[reg0 & 0x07];
    rembp = (signed char*)&vpu->byteregs[reg1 & 0x0F];
    *reswp = operand0 / operand1;
    *rembp = operand0 % operand1;
  }

  else if(reg1 & 0x80)
  {
    operand0 = *(signed char*)&vpu->byteregs[reg0 & 0x0F];
    operand1 = *(signed short*)&vpu->regs[reg1 & 0x07];
    if(!operand1)
    {
//      vpu->flags.divzero = 1;
      send_vpu_signal(vpu, SIGFPE);
      return;
    }
    resbp = (signed char*)&vpu->byteregs[reg0 & 0x0F];
    remwp = (signed short*)&vpu->regs[reg1 & 0x07];
    *resbp = operand0 / operand1;
    *remwp = operand0 % operand1;
  }

  else
  {
    operand0 = *(signed char*)&vpu->byteregs[reg0 & 0x0F];
    operand1 = *(signed char*)&vpu->byteregs[reg1 & 0x0F];
    if(!operand1)
    {
//      vpu->flags.divzero = 1;
      send_vpu_signal(vpu, SIGFPE);
      return;
    }
    resbp = (signed char*)&vpu->byteregs[reg0 & 0x0F];
    rembp = (signed char*)&vpu->byteregs[reg1 & 0x0F];
    *resbp = operand0 / operand1;
    *rembp = operand0 % operand1;
  }
  
  if(prep_overflow != is_same_sign(vpu, reg0, reg1))
    vpu->flags.overflow=1;
  vpu_set_sign(vpu, reg0);  
}
