int core_debug(struct console_struct *console);

/*
  Sends a signal to a process.
  Returns 1 if the signal could not be send. Otherwise returns 0.
*/
int send_vpu_signal(struct vpu *vpu, unsigned short signal)
{
  if(!vpu->signal || signal == SIGKILL || signal == SIGSTOP || signal == SIGCONT)
    vpu->signal = signal;
  else return 1;
  if(debug)
  {
    sprintf(message, "Signal %u to process %u", signal, vpu->pid);
    putstr(message);
  }
  return 0;
}

void process_vpu_signal(struct vpu *vpu)
{
  unsigned n;
  if(vpu->signal == SIGHIDECON || vpu->signal == SIGSHOWCON)
  {
    for(n=window_count;n--;)
      if(windows[n]->function_pointer == vpuconsole_program
      && ((struct console_struct *)windows[n]->function_status)->vcpu == vpu)
        windows[n]->hidden = SIGSHOWCON - vpu->signal;
    running_window = 0;
    set_active_window(window_visibility_order[window_count-2], 1);
    draw_screen();
  }
  else if(vpu->signal == SIGMUTLOCK)
  {
    vpu->mutex_lock = 1;
  }
  else if(vpu->signal == SIGMUTUNLOCK)
  {
    for(n=vpu_count;n--;)
      if(vpus[n]->data == vpu->data)
        vpu->mutex_lock = 0;
  }
  else switch(vpu->signal)
  {
    // default action == exit
    case SIGHUP: case SIGINT: case SIGPIPE: case SIGALRM: case SIGTERM:
    case SIGUSR1: case SIGUSR2: case SIGPOLL: case SIGVTALRM: case SIGPROF:
      if(vpu->interrupt_count > vpu->signal && vpu->interrupts[vpu->signal].address)
        goto sdefault;
    case SIGKILL:skill:
      vpu->exiting = 1;
      break;
    // default action == stop
    case SIGTSTP: case SIGTTIN: case SIGTTOU:
      if(vpu->interrupt_count > vpu->signal && vpu->interrupts[vpu->signal].address)
        goto sdefault;
    case SIGSTOP:
      vpu->pause = 1;
      break;
    case SIGCONT:
      vpu->pause = 0;
      vpu->sleep = 0;
      break;
    // default action == core
    case SIGQUIT: case SIGILL: case SIGTRAP: case SIGABRT: case SIGEMT: case SIGFPE:
    case SIGBUS: case SIGSEGV: case SIGSYS: case SIGXCPU: case SIGXFSZ:
      if(vpu->interrupt_count <= vpu->signal || !vpu->interrupts[vpu->signal].address)
      {
        sprintf(message, "Guru meditation: %u, 0x%.4X:0x%.4X", vpu->pid, vpu->code_segment, vpu->ip);
        putstr(message);
        // show core dump... if out of memory, kill process
        if(core_debug(running_console->function_status))
          goto skill;
        else
          vpu->wait=1;
      }
      goto sdefault;
    // default action == ignore
/*    case SIGCHLD:
      if(vpu->interrupt_count <= vpu->signal || !vpu->interrupts[vpu->signal].address)
        break;*/
    default:sdefault:
      vpu->interrupt_wait = vpu->signal;
      break;
  }
  vpu->signal=0;
}
