#include "sigdef.c"
#include "debug.c"
#include "unixtime.c"
#include "jump.c"
#include "printstr.c"
#include "mov.c"
#include "incdec.c"
#include "comp.c"
#include "stack.c"
#include "addsub.c"
#include "segment.c"
#include "muldiv.c"
#include "printint.c"
#include "bitwise.c"
#include "callret.c"
#include "intiret.c"
#include "vpustr.c"
#include "inout.c"
#include "wait.c"
#include "setflags.c"
#include "signconv.c"
#include "getstr.c"
#include "stacks.c"
#include "vpuwin.c"
#include "signals.c"
#include "memprot.c"
#include "int2vpu.c"
#include "syscall.c"

void (**instruction_func_ptr)(struct vpu *);

unsigned instruction_count = sizeof(INSTR) / sizeof(char*);

int init_instruction_table(void)
{
  if(!(instruction_func_ptr = malloc(instruction_count * sizeof(*instruction_func_ptr))))
    return 0;

  instruction_func_ptr[OPCODE_WAIT] = vpu_instr_wait;
  instruction_func_ptr[OPCODE_MOV] = vpu_instr_mov;
  instruction_func_ptr[OPCODE_JUMP] = vpu_instr_jump;
  instruction_func_ptr[OPCODE_INC] = vpu_instr_inc;
  instruction_func_ptr[OPCODE_DEC] = vpu_instr_dec;
  instruction_func_ptr[OPCODE_CMP] = vpu_instr_cmp;
  instruction_func_ptr[OPCODE_JE] = vpu_instr_je;
  instruction_func_ptr[OPCODE_PUSH] = vpu_instr_push;
  instruction_func_ptr[OPCODE_POP] = vpu_instr_pop;
  instruction_func_ptr[OPCODE_ADD] = vpu_instr_add;
  instruction_func_ptr[OPCODE_SUB] = vpu_instr_sub;
  instruction_func_ptr[OPCODE_DSEG] = vpu_instr_dseg;
  instruction_func_ptr[OPCODE_MUL] = vpu_instr_mul;
  instruction_func_ptr[OPCODE_DIV] = vpu_instr_div;
  instruction_func_ptr[OPCODE_PRINTUINT] = vpu_instr_printuint;
  instruction_func_ptr[OPCODE_AND] = vpu_instr_and;
  instruction_func_ptr[OPCODE_OR] = vpu_instr_or;
  instruction_func_ptr[OPCODE_XOR] = vpu_instr_xor;
  instruction_func_ptr[OPCODE_INV] = vpu_instr_inv;
  instruction_func_ptr[OPCODE_SHL] = vpu_instr_shl;
  instruction_func_ptr[OPCODE_SHR] = vpu_instr_shr;
  instruction_func_ptr[OPCODE_CALL] = vpu_instr_call;
  instruction_func_ptr[OPCODE_RET] = vpu_instr_ret;
  instruction_func_ptr[OPCODE_INT] = vpu_instr_int;
  instruction_func_ptr[OPCODE_IRET] = vpu_instr_iret;
  instruction_func_ptr[OPCODE_SETINT] = vpu_instr_setint;
  instruction_func_ptr[OPCODE_JNE] = vpu_instr_jne;
  instruction_func_ptr[OPCODE_JG] = vpu_instr_jg;
  instruction_func_ptr[OPCODE_JGE] = vpu_instr_jge;
  instruction_func_ptr[OPCODE_JLE] = vpu_instr_jle;
  instruction_func_ptr[OPCODE_JCZ] = vpu_instr_jcz;
  instruction_func_ptr[OPCODE_JL] = vpu_instr_jl;
  instruction_func_ptr[OPCODE_PRINTSTR] = vpu_instr_printstr;
  instruction_func_ptr[OPCODE_JFS] = vpu_instr_jfs;
  instruction_func_ptr[OPCODE_JFNS] = vpu_instr_jfns;
  instruction_func_ptr[OPCODE_MEMCPY] = vpu_instr_memcpy;
  instruction_func_ptr[OPCODE_IN] = vpu_instr_in;
  instruction_func_ptr[OPCODE_OUT] = vpu_instr_out;
  instruction_func_ptr[OPCODE_MEMSET] = vpu_instr_memset;
  instruction_func_ptr[OPCODE_ZFLAGS] = vpu_instr_zeroflags;
  instruction_func_ptr[OPCODE_PRINTSINT] = vpu_instr_printsint;
  instruction_func_ptr[OPCODE_IDIV] = vpu_instr_idiv;
  instruction_func_ptr[OPCODE_SCONVW] = vpu_instr_sconvw;
  instruction_func_ptr[OPCODE_SCONVB] = vpu_instr_sconvb;
  instruction_func_ptr[OPCODE_GETSTR] = vpu_instr_getstr;
  instruction_func_ptr[OPCODE_SYSCALL] = vpu_instr_sys;
  instruction_func_ptr[OPCODE_SCRCALL] = vpu_instr_scr;
  instruction_func_ptr[OPCODE_PSTACK] = vpu_select_primary_stack;
  instruction_func_ptr[OPCODE_SSTACK] = vpu_select_secondary_stack;

  return 1;
}

void vpu_console_instr(struct vpu *vpu, unsigned char *instr)
{
  unsigned char old_ip = vpu->ip;

  vpu->retwait = 0;  

  vpu->code = calloc(vpu->code_count, sizeof(char*));
  if(!vpu->code)
  {
    putstr(outofmemory);
    vpu->code = vpu->realcode;
    return;
  }

  vpu->code[vpu->code_segment] = instr;
  vpu->ip = 1;

  if(debug)
    vpu_step_debuginfo(vpu, *instr);   
  (*instruction_func_ptr[*instr])(vpu);

  free(vpu->code);
  vpu->code = vpu->realcode;

  if(*instr == OPCODE_CALL)
    vpu->stack[vpu->sp+2] = old_ip;

  if(!instr_in_array(*instr, JUMP_INSTR))
    vpu->ip = old_ip;
  else
  {
    ((struct console_struct*)running_console->function_status)->runmode = 2;
    vpu->wait = 0;
    vpu->retwait = 1;
  }
}

inline void vpu_block(struct vpu *vpu)
{
  vpu->ip--;
  steps = 0;
}

/*
  Returns -1 if no mutex set, else returns child ID that has mutex lock.
*/
inline int vpu_child_has_mutex(struct vpu *vpu)
{
  unsigned n = vpu->child_count;

  while(n--)
  {
    if(vpu->child[n]->is_clone && vpu->child[n]->mutex_lock)
      break;
  }
  
  return n;
}

unsigned vpu_turns(struct vpu *vpu)
{
  unsigned retval = vpu->priority;
  if(vpu_has_active_window(vpu))
    retval <<= 2;
  return retval;
}

inline unsigned forks_alive(struct vpu *vpu)
{
  unsigned n = vpu->child_count;

  while(n--)
    if(!vpu->child[n]->is_clone)
      return 1;
  return 0;
}

struct vpu *prepare_vpu_steps(struct vpu *vpu)
{
  int mutex;
  unsigned n=0;

  if(((struct console_struct *)running_console->function_status)->console_kill)
    return 0;  

  if(!vpu->child_count)
  {
//    vpu->child_running=-1;    
    if(vpu->exiting) return 0;
    parent:
    steps = vpu_turns(vpu);
    return vpu;
  }
   
  else
  {   
    if(vpu->mutex_lock)
      vpu->child_running = -1;

    else if((mutex = vpu_child_has_mutex(vpu)) != -1)
    {
      while(vpu->child_running == -1 || (vpu->child[vpu->child_running]->is_clone && vpu->child_running != mutex))
        if(vpu->child_running == -1) vpu->child_running = vpu->child_count-1;
        else vpu->child_running--;
    }

    else
    {
      do
        vpu->child_running--;      
      while(vpu->child_running >= (signed)vpu->child_count);
        
      if(vpu->exiting)
      {          
        if(!forks_alive(vpu))
          return 0;
          
        if(!vpu->zombie)
          free_vpu_resources(vpu);

        if(vpu->child_running < 0)
          vpu->child_running = vpu->child_count-1;          
                 
        while(vpu->child[vpu->child_running]->is_clone)
          if(--vpu->child_running < 0)
            vpu->child_running = vpu->child_count-1;
      }
      else
        if(vpu->child_running < -1)
          vpu->child_running = vpu->child_count-1;
    }

    if(vpu->child_running == -1)
      goto parent;

    steps = vpu_turns(vpu->child[vpu->child_running]);
    return vpu->child[vpu->child_running];
  }
}

unsigned char vpu_step(struct vpu *vpu);

inline struct vpu *vpu_recursive(struct vpu *vpu)
{
  struct vpu *retval;
  struct vpu *vpu2 = 0;

  while(retval = prepare_vpu_steps(vpu))
  {
    if(retval == vpu)
      return retval;
    vpu = vpu2 = retval;
  }
  if(vpu2) return vpu2;
  return 0;
}

unsigned char vpu_run(struct vpu *vpu)
{
//  if(((struct console_struct *)running_console->function_status)->console_kill)
//    return 2;
//  if(!((struct console_struct *)running_console->function_status)->runmode)
//    return 0;

  if(!(vpu = vpu_recursive(vpu)))
  {
    return 2;    
  }
  
  while(steps--)
    switch(vpu_step(vpu))
    {
      case 3: return 3;
      case 2: return 2;
      case 1: return 1;
      default: break;
    }
  return 0;
}

unsigned char vpu_step(struct vpu *vpu)
{
  unsigned instruction;
//  unsigned undead_clones = 0;

/*  int n = vpu_count;
  
  while(n--)
    if(vpu == vpus[n])
      break;
  if(n == -1)
    abort();*/

  if(vpu->exiting)
  {
//    putstr("exiting..");
    if(!vpu->child_count)
    {
      if(vpu->parent)
        delete_vpu(vpu);
      else
        return 2;      
    }
    else
    {
      send_termination_signal(vpu);
    }
    return 3;    
  }
  
  if(vpu->signal)
  {
    process_vpu_signal(vpu);
  }

  if(vpu->pause)
  {
    if(vpu->sleep && vpu->sleep < unixtime(timezone).time)
      send_vpu_signal(vpu, SIGCONT);
    return 3;
  }

  if(vpu->interrupt_wait && !vpu->interrupt_run && ((struct console_struct *)running_console->function_status)->runmode)
  {
//    putstr("signal");
//    steps=0;
    if(vpu->interrupt_wait < vpu->interrupt_count)
    {
      vpu->wait = 0;      
      vpu->interrupt_run=1;      
      vpu_instr_hardint(vpu, vpu->interrupt_wait);
    }
    vpu->interrupt_wait = 0;
    return 3;
  }
  
  if(vpu->wait)
  {
    vpu_getwindowkeys(vpu);
    return 3;
  }

  if(vpu->ip >= vpu->code_size)
  {
    send_vpu_signal(vpu, SIGSEGV);
    sprintf(message, "PID %u: Code offset violation!", vpu->pid);
    putstr(message);
    return 0;
  }
  else instruction = vpu->code[vpu->code_segment][vpu->ip++];

  if(debug)
  {
    sprintf(message, "PID: %u, Instruction pointer: 0x%.4X, Instruction opcode: 0x%.2X", vpu->pid, vpu->ip-1, instruction);
    putstr(message);
  }

  if(instruction > instruction_count)
    send_vpu_signal(vpu, SIGILL);

  else
  {
    if(debug)
      vpu_step_debuginfo(vpu, instruction);
    (*instruction_func_ptr[instruction])(vpu);
    vpu_getwindowkeys(vpu);
  }

  return 0;
}

