#define OPEN_O_READ 0x0000
#define OPEN_O_WRITE 0x0001
#define OPEN_O_APPEND 0x0002
#define OPEN_O_RW 0x0004
#define OPEN_O_SYNC 0x0008
#define OPEN_O_NONBLOCK 0x1000
#define PIPE_BUF 0x200

#include "biosser.c"
#include "biospar.c"

int load_driver(char *program, char *modname);
unsigned char device_exists(unsigned driver, unsigned device);
int driver_id(char *name);
extern struct driver *drivers;
char modfs[] = "DEV:\\";
int dev_cmd(unsigned driver, int device, void __far *cmd);
int dev_open(unsigned driver, unsigned device, unsigned mode);
int dev_write(unsigned driver, unsigned device, unsigned char byte);
int dev_close(unsigned driver, unsigned device);
int dev_read(unsigned driver, unsigned device);
int dev_fsync(unsigned driver, int device);

void sync_dup(struct file_struct *fd0, struct file_struct *fd1)
{
//  sprintf(message, "Synkataan dupit %p ja %p", fd0, fd1);
//  putstr(message);
  
  if(fd0->file_pos && fd1->file_pos)
    *fd0->file_pos = *fd1->file_pos;
  fd0->read_ahead = fd1->read_ahead;
  fd0->write_buff = fd1->write_buff;
  fd0->pipe_from = fd1->pipe_from;
  fd0->pipe_to = fd1->pipe_to;
  fd0->flags = fd1->flags;
  fd0->bufrpos = fd1->bufrpos;
  fd0->bufwpos = fd1->bufwpos;
}

void fd_dup_sync(struct file_struct *fd)
{
  struct file_struct *dupfd = fd;

  while(dupfd = dupfd->dup)
  {
    sync_dup(dupfd, fd);
  }

  dupfd = fd;
  while(dupfd = dupfd->orig)
  {
    sync_dup(dupfd, fd);
  }
}

void dd_dup_sync(struct dir_struct *dd)
{
 // todo...
}

int pipe_read(struct file_struct *pipe)
{
  int c;

//  sprintf(message, "Luetaan pipest %p puskuri %p", pipe, pipe->buff);
//  putstr(message);  

  if(pipe->bufrpos != pipe->bufwpos)
  {
    c = pipe->buff[pipe->bufrpos++ & 0x1FF];
    fd_dup_sync(pipe);
    return c;
  }
  return -1;
}

int con_stdin_read(void)
{
  if(!strchr(((struct console_struct*)running_console->function_status)->con_stdin.buff, 13)
    && !strchr(((struct console_struct*)running_console->function_status)->con_stdin.buff, 10))
    return -1;

  if(((struct console_struct*)running_console->function_status)->con_stdin.buff
    [((struct console_struct*)running_console->function_status)->con_stdin.pos])
    return ((struct console_struct*)running_console->function_status)->con_stdin.buff
      [((struct console_struct*)running_console->function_status)->con_stdin.pos++];

  *((struct console_struct*)running_console->function_status)->con_stdin.buff = 0;
  ((struct console_struct*)running_console->function_status)->con_stdin.pos = 0;
  ((struct console_struct*)running_console->function_status)->input_col = 0;

  if(((struct console_struct*)running_console->function_status)->con_stdin.len != 32)
  {
    ((struct console_struct*)running_console->function_status)->con_stdin.buff
      = realloc(((struct console_struct*)running_console->function_status)->con_stdin.buff, 32*sizeof(char));
    ((struct console_struct*)running_console->function_status)->con_stdin.len = 32;      
  }
  
  return -1;
}


int con_stdout_write(unsigned char c)
{
  void *new_pointer;
  
  ((struct console_struct*)running_console->function_status)->con_stdout.buff
    [((struct console_struct*)running_console->function_status)->con_stdout.pos++] = c;

  ((struct console_struct*)running_console->function_status)->con_stdout.buff
    [((struct console_struct*)running_console->function_status)->con_stdout.pos] = 0;  

  if(((struct console_struct*)running_console->function_status)->con_stdout.pos
    == ((struct console_struct*)running_console->function_status)->con_stdout.len-1)
  {
    new_pointer = realloc(((struct console_struct*)running_console->function_status)->con_stdout.buff,
      (((struct console_struct*)running_console->function_status)->con_stdout.len += 32) * sizeof(char));

    if(!new_pointer)
      return 2;
    ((struct console_struct*)running_console->function_status)->con_stdout.buff = new_pointer;
  }

  if(/*c == 13 || */c == 10)
  {
    putstr(((struct console_struct*)running_console->function_status)->con_stdout.buff);
    if(((struct console_struct*)running_console->function_status)->con_stdout.len != 32)
    {
      ((struct console_struct*)running_console->function_status)->con_stdout.buff
        = realloc(((struct console_struct*)running_console->function_status)->con_stdout.buff, 32*sizeof(char));
      ((struct console_struct*)running_console->function_status)->con_stdout.len = 32;
    }
    ((struct console_struct*)running_console->function_status)->con_stdout.pos = 0;
    *((struct console_struct*)running_console->function_status)->con_stdout.buff = 0;
    return 0;      
  }


  return 0;
}

void vpu_syscall_read(struct vpu *vpu)
{
  unsigned char *buf = &vpu->data[vpu->data_segment][vpu->regs[1]];
  unsigned short count = vpu->regs[2];
  unsigned short n = 0;
  int c;

  if(!memory_allowed(vpu, vpu->regs[1] + count))
  {
    vpu->regs[0] = 0xFFFF;
    return;
  }

  if(vpu->regs[3] < MAX_FILES && vpu->file_handles[vpu->regs[3]])
  {
    if(!vpu->file_struct[vpu->regs[3]]->is_ioport && vpu->file_struct[vpu->regs[3]]->write_buff)
    {
      vpu->regs[0] = fflush(vpu->file_handles[vpu->regs[3]]);
      if(vpu->regs[0])
        return;
      vpu->file_struct[vpu->regs[3]]->write_buff = 0;
    }
    while(count--)    
    {
      if(vpu->file_struct[vpu->regs[3]]->is_ioport)
      {
        if(vpu->file_struct[vpu->regs[3]]->is_ioport == 9)
          c = dev_read(vpu->file_struct[vpu->regs[3]]->module, vpu->file_struct[vpu->regs[3]]->device);
        else if(vpu->file_struct[vpu->regs[3]]->is_ioport <= 4)
          c = serial_read(vpu->file_struct[vpu->regs[3]]->is_ioport-1);
        else if(vpu->file_struct[vpu->regs[3]]->is_ioport == 11)
          c = pipe_read(vpu->file_struct[vpu->regs[3]]);
        else if(vpu->file_struct[vpu->regs[3]]->is_ioport == 12)
        {
          c = con_stdin_read();
        }
        else// if(vpu->file_struct[vpu->regs[3]]->is_ioport > 4)
          goto minus1;

        if(c == -1)
        {
          if(!n && !(vpu->file_struct[vpu->regs[3]]->flags & OPEN_O_NONBLOCK))
          {
//            vpu->ip--;
//            steps=0;
            vpu_block(vpu);
            return;
          }          
          break;
        }
        else if(c == -2)
        {
          minus1:n = 0xFFFF;
          break;
        }
      }      
      else
      {
        c = fgetc(vpu->file_handles[vpu->regs[3]]);
        if(c == EOF)
        {
          vpu->file_struct[vpu->regs[3]]->read_ahead = 0;
          break;
        }
      }
      buf[n++] = c;
//      sprintf(message, "read to %u", vpu->regs[1]+n);
//      putstr(message);
      if(vpu->file_struct[vpu->regs[3]]->file_pos)
        ++*vpu->file_struct[vpu->regs[3]]->file_pos;
    }
    if(c != EOF)
      vpu->file_struct[vpu->regs[3]]->read_ahead += n;
  }
  vpu->regs[0] = n;
  fd_dup_sync(vpu->file_struct[vpu->regs[3]]);  
}

int pipe_write(struct vpu *vpu, struct file_struct *pipe, unsigned char chr)
{
  if(!pipe->pipe_to)
  {
    if(send_vpu_signal(vpu, SIGPIPE))
      send_vpu_signal(vpu, SIGABRT);
    return 2;
  }
//  sprintf(message, "Kirjoitetaan pipeen %p > %p puskuri %p kahva", pipe, pipe->pipe_to, pipe->pipe_to->buff);
//  putstr(message);    
  if(((pipe->pipe_to->bufwpos+1) & 0x1FF) == (pipe->pipe_to->bufrpos & 0x1FF))
  {
    return 1;
  }
  pipe->pipe_to->buff[pipe->pipe_to->bufwpos++ & 0x1FF] = chr;

  fd_dup_sync(pipe->pipe_to);

  return 0;
}

void vpu_syscall_write(struct vpu *vpu)
{
  unsigned char *buf = &vpu->data[vpu->data_segment][vpu->regs[1]];
  unsigned short count = vpu->regs[2];
  unsigned short n = 0;
  unsigned char r;

  if(!memory_allowed(vpu, vpu->regs[1] + count)) goto error;

//  sprintf(message, "Kirjoitetaan %u-deskriptorin tiedostoon", vpu->file_handles[vpu->regs[3]]);
//  putstr(message);

  if(vpu->regs[3] < MAX_FILES && vpu->file_handles[vpu->regs[3]])
  {
//    sprintf(message, "Kirjoitetaan %u-tyyppiseen tiedostoon", vpu->file_struct[vpu->regs[3]]->is_ioport);
//    putstr(message);
    if(!vpu->file_struct[vpu->regs[3]]->is_ioport && vpu->file_struct[vpu->regs[3]]->read_ahead)
    {
      fseek(vpu->file_handles[vpu->regs[3]], vpu->file_struct[vpu->regs[3]]->read_ahead, SEEK_CUR);
      vpu->file_struct[vpu->regs[3]]->read_ahead = 0;
    }
    while(count--)
    {     
      if(vpu->file_struct[vpu->regs[3]]->is_ioport)
      {
        if(vpu->file_struct[vpu->regs[3]]->is_ioport <= 4)
          r = serial_write(vpu->file_struct[vpu->regs[3]]->is_ioport-1, buf[n]);
        else if(vpu->file_struct[vpu->regs[3]]->is_ioport < 9)
          r = parallel_write(vpu->file_struct[vpu->regs[3]]->is_ioport-5, buf[n]);
        else if(vpu->file_struct[vpu->regs[3]]->is_ioport == 9)
          r = dev_write(vpu->file_struct[vpu->regs[3]]->module, vpu->file_struct[vpu->regs[3]]->device, buf[n]);
        else if(vpu->file_struct[vpu->regs[3]]->is_ioport == 10)
          r = pipe_write(vpu, vpu->file_struct[vpu->regs[3]], buf[n]);
        else if(vpu->file_struct[vpu->regs[3]]->is_ioport == 12)
          r = con_stdout_write(buf[n]);
        switch (r)
        {
          //                  error levels:
          case 0: break; // no error, continue writing
          case 1: //        out of paper or some other non-critical error that prevents writing, block or return count of bytes written
            if(!n && !(vpu->file_struct[vpu->regs[3]]->flags & OPEN_O_NONBLOCK))
            {
//              vpu->ip--;
//              steps=0;
              vpu_block(vpu);
              return;
            }
            else
              goto end;
          case 2: //        catastroof happenedz, return -1
            goto error;
        }
      }
      else if(fputc(buf[n], vpu->file_handles[vpu->regs[3]]) == EOF)
      {
        error:vpu->regs[0] = 0xFFFF;
        return;
      }
      n++;
      if(vpu->file_struct[vpu->regs[3]]->file_pos)
        ++*vpu->file_struct[vpu->regs[3]]->file_pos;
    }
    if(!vpu->file_struct[vpu->regs[3]]->is_ioport && vpu->file_struct[vpu->regs[3]]->flags & OPEN_O_SYNC)
    {
      vpu->regs[0] = fflush(vpu->file_handles[vpu->regs[3]]);
      goto end;
    }
    else
      vpu->file_struct[vpu->regs[3]]->write_buff += n;
  }
  end:vpu->regs[0] = n;
  fd_dup_sync(vpu->file_struct[vpu->regs[3]]);
}

void vpu_syscall_fsync(struct vpu *vpu)
{
  if(vpu->regs[1] < MAX_FILES && vpu->file_handles[vpu->regs[1]])
  {
    if(!vpu->file_struct[vpu->regs[1]]->is_ioport)
    {
      if(!(vpu->regs[0] = fflush(vpu->file_handles[vpu->regs[1]])))
        vpu->file_struct[vpu->regs[1]]->write_buff = 0;
    }
    else if(vpu->file_struct[vpu->regs[1]]->is_ioport == 9)
    {
      dev_fsync(vpu->file_struct[vpu->regs[3]]->module, vpu->file_struct[vpu->regs[3]]->device);
    }
  }
  fd_dup_sync(vpu->file_struct[vpu->regs[1]]);  
}

char *openflags2fopenmod(unsigned short flags)
{
  static char retval[4];

  retval[0] = 'r';
  retval[1] = 'b';
  retval[2] = 0;
  retval[3] = 0;  

  if(flags & OPEN_O_WRITE)
    retval[0] = 'w';
  if(flags & OPEN_O_APPEND)
    retval[0] = 'a';    
  if(flags & OPEN_O_RW)
    strcat(retval, "+");

  return retval;
}

unsigned char detect_ioport(char *filename)
{
  if(strlen(filename) == 4 && filename[3] > 0x30 && filename[3] < 0x35)
  {
    if(!strncmp(filename, "COM", 3))
      return filename[3] - 0x30;
    if(!strncmp(filename, "LPT", 3))
      return filename[3] - 0x30+4;      
  }
  else if(!strcmp(filename, "CON"))
    return 12;
  return 0;
}

void vpu_syscall_open(struct vpu *vpu)
{
  char *filename = &vpu->data[vpu->data_segment][vpu->regs[3]];
//  unsigned short filemode = vpu->data[vpu->data_segment][vpu->regs[2]]; // permissions...
  unsigned short n, m;
  unsigned short file_descriptor;
  char devname[6];
  unsigned char devnum;
  char *fileflags = openflags2fopenmod(vpu->regs[1]);

  for(n=0;filename[n];n++)
    if(!memory_allowed(vpu, vpu->regs[3] + n))
      return;
  if(!n)
    return;
      

  for(file_descriptor=0;vpu->file_handles[file_descriptor] && file_descriptor < MAX_FILES;file_descriptor++);

  
  if(file_descriptor < MAX_FILES)
  {
    if(vpu->file_struct[file_descriptor]->is_ioport = detect_ioport(filename))
    {
      if(vpu->file_struct[file_descriptor]->is_ioport <= 4)
      {
        if(serial_init(vpu->file_struct[file_descriptor]->is_ioport-1, vpu->regs[1]))
          goto error;
      }          
      else if(vpu->file_struct[file_descriptor]->is_ioport <= 8)
      {
        if(parallel_init(vpu->file_struct[file_descriptor]->is_ioport-5, vpu->regs[1]))
          goto error;
      }
      vpu->file_handles[file_descriptor] = (FILE *)-1;        
      goto ioport;
    }
    else if(!strncmp(modfs, filename, strlen(modfs)))
    {
      strncpy(devname, &filename[strlen(modfs)], 5);      
      if(driver_id(devname) < 0)
      {
        m = strlen(devname)-1;
//        devname = &filename[strlen(modfs)];
        devnum = devname[m] - 0x30;
        devname[m] = 0;
      }
      if((vpu->file_struct[file_descriptor]->module = driver_id(devname)) < 0)
        goto error;        
      vpu->file_struct[file_descriptor]->is_ioport = 9;
      vpu->file_struct[file_descriptor]->device = devnum;
      if(dev_open(vpu->file_struct[file_descriptor]->module, devnum, vpu->regs[1]))
        goto error;     
      vpu->file_handles[file_descriptor] = (FILE *)-1;
      goto ioport; 
    }
    else
    {
      vpu_cwd(vpu);
      vpu->file_handles[file_descriptor] = fopen(filename, fileflags);
    }
    if(!vpu->file_handles[file_descriptor])
    {
      error:vpu->regs[0] = 0xFFFF;
      return;
    }
    if(vpu->regs[4] && memory_allowed(vpu, vpu->regs[4] + sizeof(unsigned long)-1))
    {
      vpu->file_struct[file_descriptor]->file_pos = (unsigned long*)&vpu->data[vpu->data_segment][vpu->regs[4]];
      vpu->file_struct[file_descriptor]->filepos_segment = vpu->data_segment;
      vpu->file_struct[file_descriptor]->filepos_offset = vpu->regs[4];
      *vpu->file_struct[file_descriptor]->file_pos = ftell(vpu->file_handles[file_descriptor]);
    }
//    else
      ioport://vpu->file_struct[file_descriptor]->file_pos = 0;
    vpu->regs[0] = file_descriptor;
    vpu->file_struct[file_descriptor]->flags = vpu->regs[1];

//    sprintf(message, "Opened file descriptor %u", file_descriptor);
//    putstr(message);
  }
  else
    vpu->flags.datav = 1;
}

void vpu_fd_close(struct vpu *vpu, unsigned fd)
{
//    sprintf(message, "PID %u Closing FD %u type %u", vpu->pid, fd, vpu->file_struct[vpu->regs[1]]->is_ioport);
//    putstr(message);    
  if(vpu->file_struct[fd]->dup || vpu->file_struct[fd]->orig)
  {    
    if(vpu->file_struct[fd]->orig)
    {
      if(vpu->file_struct[fd]->pipe_from)
      {
        vpu->file_struct[fd]->pipe_from->pipe_to = vpu->file_struct[fd]->orig;
        fd_dup_sync(vpu->file_struct[fd]->pipe_from);
      }
      if(vpu->file_struct[fd]->pipe_to)
      {
        vpu->file_struct[fd]->pipe_to->pipe_from = vpu->file_struct[fd]->orig;
        fd_dup_sync(vpu->file_struct[fd]->pipe_to);
      }
      vpu->file_struct[fd]->orig->dup = vpu->file_struct[fd]->dup;
    }
    if(vpu->file_struct[fd]->dup)
    {
      if(vpu->file_struct[fd]->pipe_from)
      {
        vpu->file_struct[fd]->pipe_from->pipe_to = vpu->file_struct[fd]->dup;
        fd_dup_sync(vpu->file_struct[fd]->pipe_from);        
      }
      if(vpu->file_struct[fd]->pipe_to)
      {
        vpu->file_struct[fd]->pipe_to->pipe_from = vpu->file_struct[fd]->dup;
        fd_dup_sync(vpu->file_struct[fd]->pipe_to);
      }      
      vpu->file_struct[fd]->dup->orig = vpu->file_struct[fd]->orig;
    }
    goto zeroall;
  }
  
  else
  {
    if(vpu->file_struct[fd]->is_ioport)
    {
      if(vpu->file_struct[fd]->is_ioport <= 4)
        close_serialport(vpu->file_struct[fd]->is_ioport -1);
      else if(vpu->file_struct[fd]->is_ioport == 10) // close pipe's write end
      {
        if(vpu->file_struct[fd]->pipe_to)
        {
//          sprintf(message, "Suljetaan kirjoitusputki %p > %p puskuri %p", vpu->file_struct[fd], vpu->file_struct[fd]->pipe_to, vpu->file_struct[fd]->pipe_to->buff);
//          putstr(message);                              
          vpu->file_struct[fd]->pipe_to->pipe_from = 0;
          vpu->file_struct[fd]->pipe_to->flags |= OPEN_O_NONBLOCK; // make read end non-blocking
          fd_dup_sync(vpu->file_struct[fd]->pipe_to);
        }
      }
      else if(vpu->file_struct[fd]->is_ioport == 11) // close pipe's read end
      {
//        sprintf(message, "Suljetaan lukuputki %p puskuri %p", vpu->file_struct[fd], vpu->file_struct[fd]->buff);
//        putstr(message);            
        free(vpu->file_struct[fd]->buff);
        if(vpu->file_struct[fd]->pipe_from)
        {
          vpu->file_struct[fd]->pipe_from->pipe_to = 0;
          fd_dup_sync(vpu->file_struct[fd]->pipe_from);
        }
      }
//      else if(vpu->file_struct[fd]->is_ioport == 12); // close console io
      else if(vpu->file_struct[fd]->is_ioport == 9)
        if(dev_close(vpu->file_struct[fd]->module, vpu->file_struct[fd]->device))
          goto error;
      
      goto zeroall;
    }
    else if(!fclose(vpu->file_handles[fd]))
    {
      zeroall:
      vpu->regs[0] = 0;
      vpu->file_handles[fd] = 0;
      memset(vpu->file_struct[fd], 0, sizeof(struct file_struct));
    }
    else
      error:vpu->regs[0] = 0xFFFF;
  }
}

void vpu_syscall_close(struct vpu *vpu)
{
  if(vpu->regs[1] < MAX_FILES && vpu->file_handles[vpu->regs[1]])
  {
//    sprintf(message, "PID %u Closing FD %u type %u", vpu->pid, vpu->regs[1], vpu->file_struct[vpu->regs[1]]->is_ioport);
//    putstr(message);
    vpu_fd_close(vpu, vpu->regs[1]);
  }
  else
    vpu->flags.datav = 1;
}

void vpu_syscall_lseek(struct vpu *vpu)
{
  if(vpu->regs[3] < MAX_FILES && vpu->file_handles[vpu->regs[3]] && !vpu->file_struct[vpu->regs[3]]->is_ioport)
  {
    if(vpu->file_struct[vpu->regs[3]]->read_ahead)
    {
      fseek(vpu->file_handles[vpu->regs[3]], vpu->file_struct[vpu->regs[3]]->read_ahead, SEEK_CUR);
      vpu->file_struct[vpu->regs[3]]->read_ahead = 0;
    }
    switch(vpu->regs[2])
    {
      case 0:
        vpu->regs[0] = fseek(vpu->file_handles[vpu->regs[3]], *(signed short*)&vpu->regs[1], SEEK_SET);
        break;
      case 1:
        vpu->regs[0] = fseek(vpu->file_handles[vpu->regs[3]], *(signed short*)&vpu->regs[1], SEEK_CUR);
        break;
      case 2:
        vpu->regs[0] = fseek(vpu->file_handles[vpu->regs[3]], *(signed short*)&vpu->regs[1], SEEK_END);
    }
    if(vpu->file_struct[vpu->regs[3]])
      *vpu->file_struct[vpu->regs[3]]->file_pos = ftell(vpu->file_handles[vpu->regs[3]]);
    fd_dup_sync(vpu->file_struct[vpu->regs[3]]);      
    vpu->regs[0] = ftell(vpu->file_handles[vpu->regs[3]]);
  }
}

int vpu_allocate_new_segment(struct vpu *vpu)
{
  void *new_pointer;

  if(!vpu->dynsegment_count)
  {
    if(!(vpu->dynsegments = calloc(++vpu->dynsegment_count, sizeof(struct dynsegment))))
      return -1;
  }
  else
    if(new_pointer = realloc(vpu->dynsegments, ++vpu->dynsegment_count * sizeof(struct dynsegment)))
      vpu->dynsegments = new_pointer;
    else return -1;
      
  vpu->dynsegments[vpu->dynsegment_count-1].segment = vpu->regs[1];
  vpu->dynsegments[vpu->dynsegment_count-1].length = vpu->regs[2];
  
//  sprintf(message, "Allocating dynamic segment %u", vpu->dynsegment_count-1);
//  putstr(message);
  
  if(vpu->dataseg_table_len < vpu->regs[1]+1) 
    if(new_pointer = realloc(vpu->data, (vpu->regs[1]+1) * sizeof(unsigned char*)))
    {
      vpu->data = new_pointer;
      vpu->dataseg_table_len = vpu->regs[1]+1;
    }
    else goto outofmemory;
    
//  sprintf(message, "Data segment table is now %u long.", vpu->dataseg_table_len);
//  putstr(message);
//  sprintf(message, "%u dynamically allocated segments.", vpu->dynsegment_count);
//  putstr(message);  
  
  vpu->data[vpu->regs[1]] = calloc(vpu->regs[2], sizeof(unsigned char));
  if(!vpu->data[vpu->regs[1]])
  {
    outofmemory:putstr(outofmemory);
    return -1;
  }  
  
  return 0;
}

unsigned char vpu_segment_in_use(struct vpu *vpu, unsigned short segment)
{
  unsigned short n;
  
  if(segment < vpu->data_count)
    return 1;
  n = vpu->dynsegment_count;
  while(n--)
  {  
    if(vpu->dynsegments[n].segment == segment)
      return 1;
  }
  
  return 0;
}

void vpu_syscall_mmap(struct vpu *vpu)
{
  unsigned char *fpreadp;
  if(!vpu->regs[2])
  {
    putstr("Cannot allocate zero bytes!");
    return;
  }
  // find free segment if segment [bx] already used
  if(vpu_segment_in_use(vpu, vpu->regs[1]))
  {
    sprintf(message, "Segment %u already allocated!", vpu->regs[1]);
    putstr(message);
    for(vpu->regs[1]=vpu->data_count; vpu_segment_in_use(vpu, vpu->regs[1]) ;vpu->regs[1]++)
    {
      if(vpu->regs[1] == 0xFFFF)
      {
        putstr("Out of free memory segments!");
        goto end;
      }
    }
  }
  if(vpu_allocate_new_segment(vpu))
  {
    free(vpu->data[vpu->regs[1]]);
    goto end;
  }
  // if file descriptor argument is set, read file contents into newly allocated memory
  if(vpu->regs[4] && vpu->file_handles[vpu->regs[4]])
  {
    fpreadp = vpu->data[vpu->regs[1]];
    fseek(vpu->file_handles[vpu->regs[4]], *(signed short*)&vpu->regs[5], SEEK_SET);
    while(vpu->regs[2]-- && !feof(vpu->file_handles[vpu->regs[4]]))
      *fpreadp++ = fgetc(vpu->file_handles[vpu->regs[4]]);
    // save file position to file struct
    if(vpu->file_struct[vpu->regs[4]]->file_pos)
      *vpu->file_struct[vpu->regs[4]]->file_pos = ftell(vpu->file_handles[vpu->regs[4]]);
    fd_dup_sync(vpu->file_struct[vpu->regs[4]]);
  }
  end:vpu->regs[0]=vpu->regs[1];
}

void vpu_free_dynamic_segment(struct vpu *vpu, unsigned short segment)
{
  free(vpu->data[vpu->dynsegments[segment].segment]);
  vpu->dynsegment_count--;    
  if(segment < vpu->dynsegment_count)
    do
    {
      vpu->dynsegments[segment] = vpu->dynsegments[segment+1];
    }
    while(++segment < vpu->dynsegment_count-1);
  vpu->dynsegments = realloc(vpu->dynsegments, vpu->dynsegment_count*sizeof(struct dynsegment));
}

void vpu_syscall_munmap(struct vpu *vpu)
{
  unsigned short n = vpu->dynsegment_count;
  // cannot free current segment...
  if(vpu->data_segment == vpu->regs[1])
  {
    vpu->regs[0] = 0xFFFF;
    return; 
  }
  while(n--)
    if(vpu->dynsegments[n].segment == vpu->regs[1])
    {
      vpu_free_dynamic_segment(vpu, n);
      vpu->regs[0] = 0;
      return;
    }
  vpu->regs[0] = n;
}

void vpu_syscall_exit(struct vpu *vpu)
{
  if(!vpu->parent->zombie && vpu->exit_status)
    *vpu->exit_status = vpu->regs[1];
  vpu->exiting = 1;
}

// returns argument count to AX
void vpu_syscall_getargc(struct vpu *vpu)
{
  vpu->regs[0] = vpu->argc;
}

// returns offset [CX] of arg [BX] to AX
void vpu_syscall_getargv(struct vpu *vpu)
{
  if(vpu->regs[1] < vpu->argc && vpu->regs[2] <= strlen(vpu->argv[vpu->regs[1]]))
  {
    vpu->regs[0] = vpu->argv[vpu->regs[1]][vpu->regs[2]];
  }
}

// spawns new process
void *run_program(char *dir, char *program, char *args, unsigned getkey);
void vpu_syscall_spawn(struct vpu *vpu)
{
  char *path = &vpu->data[vpu->data_segment][vpu->regs[2]];
  char *args = &vpu->data[vpu->data_segment][vpu->regs[3]];
  unsigned *pid = (unsigned*)&vpu->data[vpu->data_segment][vpu->regs[1]];
  struct console_struct *return_console;
  unsigned n,m;
  
  for(n=0;path[n] && memory_allowed(vpu, vpu->regs[2] + n);n++);
  for(m=0;args[m] && memory_allowed(vpu, vpu->regs[2] + m);m++);  
    
  if(!memory_allowed(vpu, vpu->regs[1] + sizeof(unsigned short)-1) || !memory_allowed(vpu, vpu->regs[2]+n) || !memory_allowed(vpu, vpu->regs[3]+m))
    return;

  vpu_cwd(vpu);  
  
  return_console = run_program(vpu->cwd, path, args, 0);
  if(!return_console)
  {
    vpu->regs[0] = 0xFFFF;
    return;
  }
  *pid = return_console->vcpu->pid;
  vpu->regs[0] = 0;
}

void vpu_fd_makedup(struct vpu *vpu0, unsigned fd0, struct vpu *vpu1, unsigned fd1)
{
  struct file_struct *orig;

  vpu0->file_handles[fd0] = vpu1->file_handles[fd1];

/*  if(vpu1->file_struct[fd1]->is_ioport == 12)
  {
    vpu0->file_struct[fd0]->is_ioport = 12;
    return;
  }*/

  orig = vpu1->file_struct[fd1];
  while(orig->dup)
    orig = orig->dup;
  orig->dup = vpu0->file_struct[fd0];

  memcpy(vpu0->file_struct[fd0], vpu1->file_struct[fd1], sizeof(struct file_struct));  
  vpu0->file_struct[fd0]->orig = orig;
  vpu0->file_struct[fd0]->dup = 0;
  if(vpu0->file_struct[fd0]->file_pos)
    vpu0->file_struct[fd0]->file_pos
      = (unsigned long*)&vpu0->data[vpu0->file_struct[fd0]->filepos_segment][vpu0->file_struct[fd0]->filepos_offset];
}

inline unsigned char copy_vpu(struct vpu *vpu0, struct vpu *vpu1)
{
  unsigned n;
  void *new_pointer;

  vpu0->code = 0;
  vpu0->data = 0;
  vpu0->interrupts = 0;
  vpu0->data_count=0;
  vpu0->code_count=0;
  vpu0->dynsegments = 0;
  vpu0->dynsegment_count=0;
  vpu0->file_handles=0;
  vpu0->file_struct=0;
  vpu0->dir_handles=0;
  vpu0->ports=0;
  vpu0->stack = vpu0->secondary_stack = vpu0->primary_stack = 0;
  if(!init_vpu(vpu0, vpu1->stack_size, vpu1->code_size, vpu1->code_count, vpu1->data_size, vpu1->data_count, vpu1->interrupt_count))
  {
    return 0;
  }

  vpu0->psp = vpu1->psp;
  vpu0->ssp = vpu1->ssp;
  vpu0->sp = vpu1->sp;  

  if(vpu1->stack == vpu1->secondary_stack)
    vpu0->stack = vpu0->secondary_stack;

  memcpy(vpu0->primary_stack, vpu1->primary_stack, vpu1->stack_size*sizeof(unsigned short));
  memcpy(vpu0->secondary_stack, vpu1->secondary_stack, vpu1->stack_size*sizeof(unsigned short));  
  for(n=vpu1->code_count;n--;)
    memcpy(vpu0->code[n], vpu1->code[n], vpu1->code_size);
  for(n=vpu1->data_count;n--;)
    memcpy(vpu0->data[n], vpu1->data[n], vpu1->data_size);
  memcpy(vpu0->interrupts, vpu1->interrupts, vpu1->interrupt_count*sizeof(struct vpu_interrupt));

  if(vpu1->dynsegment_count)
  {
    vpu0->dynsegments = calloc(vpu1->dynsegment_count, sizeof(struct dynsegment));
    if(!vpu0->dynsegments)
      return 0;
    vpu0->dataseg_table_len = vpu1->dataseg_table_len;
    new_pointer = realloc(vpu0->data, vpu0->dataseg_table_len*sizeof(char *));
    if(!new_pointer)
      return 0;
    vpu0->data = new_pointer;
    while(vpu0->dynsegment_count < vpu1->dynsegment_count)
    {
      vpu0->dynsegments[vpu0->dynsegment_count] = vpu1->dynsegments[vpu0->dynsegment_count];
      vpu0->data[vpu0->dynsegments[vpu0->dynsegment_count].segment] = calloc(vpu0->dynsegments[vpu0->dynsegment_count].length, sizeof(unsigned char));
      if(!vpu0->data[vpu0->dynsegments[vpu0->dynsegment_count].segment])
        return 0;
      vpu0->dynsegment_count++;
    }
  }

  for(n=MAX_FILES;n--;)
  {
    if(vpu1->file_handles[n])
    {
      vpu_fd_makedup(vpu0, n, vpu1, n);
    }
  }
  for(n=MAX_DIRS;n--;)
  {
    *vpu0->dir_handles[n] = *vpu1->dir_handles[n];
    vpu0->dir_handles[n]->orig = vpu1->dir_handles[n];
    vpu1->dir_handles[n]->dup = vpu0->dir_handles[n];
  }

  
  return 1;
}

unsigned vpu_assign_child(struct vpu *child, struct vpu *parent, unsigned term_sig, short *retval_address)
{
  void *new_pointer;
  if(!parent->child_count++)
  {
    parent->child = calloc(parent->child_count, sizeof(struct vpu*));
    if(!parent->child)
      goto nomemory;
  }
  else
  {    
    new_pointer = realloc(parent->child, parent->child_count * sizeof(struct vpu*));
    if(!new_pointer)
    {
      nomemory:
      parent->child_count--;
      return 0;
    }
    else
      parent->child = new_pointer;
  }
  parent->child[parent->child_count-1] = child;
  child->parent = parent;
  child->termination_signal = term_sig;
  child->exit_status = retval_address;  
  return 1;
}

struct vpu *clone_vpu(struct vpu *parent)
{
  struct vpu *clone = spawn_new_vpu();
  unsigned pid;

  if(!clone)
  {
   nomemory();
   return 0;
  }
  pid = clone->pid;
  // copy all registers, stack pointer etc.  
  *clone = *parent;
  // reserve pid
  clone->pid = pid;
  clone->child = 0;
  clone->child_count = 0;
//  clone->child_running = -1;
  clone->parent=0;
  return clone;
}

void vpu_syscall_fork(struct vpu *vpu)
{
  struct vpu *forked_clone = clone_vpu(vpu);
  unsigned n;

  if(!forked_clone)
  {
    outofmemory:
    vpu->regs[0] = 0xFFFF;
    return;
  }
  forked_clone->is_clone = 0;  
  
  if(!copy_vpu(forked_clone, vpu) || !vpu_assign_child(forked_clone, vpu, vpu->regs[3], (short *)&vpu->data[vpu->data_segment][vpu->regs[4]]))
  {
    nomemory();
    delete_vpu(forked_clone);
    goto outofmemory;
  }

  vpu->regs[0] = forked_clone->pid;
  forked_clone->regs[0] = 0;
}

void vpu_syscall_clone(struct vpu *vpu)
{
  struct vpu *clone = clone_vpu(vpu);

  if(!clone)
  {
    outofmemory:
    vpu->regs[0] = 0xFFFF;
    return;
  }
  clone->is_clone=1;    
  // allocate own stack for the child
  clone->primary_stack = calloc(clone->stack_size, sizeof(unsigned short));
  clone->secondary_stack = calloc(clone->stack_size, sizeof(unsigned short));
  if(!clone->primary_stack || !clone->secondary_stack || !vpu_assign_child(clone, vpu, vpu->regs[3], (short *)&vpu->data[vpu->data_segment][vpu->regs[4]]))
  {
    nomemory();
    delete_vpu(clone);
    goto outofmemory;
  }
  if(vpu->stack == vpu->primary_stack)
    clone->stack = clone->primary_stack;
  else
    clone->stack = clone->secondary_stack;
      
  clone->psp = clone->ssp = clone->sp = clone->stack_size-1;

  clone->code_segment = vpu->regs[1];
  clone->ip = vpu->regs[2];
 
  vpu->regs[0] = clone->pid;
}

inline void remove_fd_filepos_ptrs(struct vpu *vpu)
{
  unsigned n = MAX_FILES;
  while(n--)
  {
    vpu->file_struct[n]->file_pos = 0;
  }
}

char **split_args(unsigned *argc, char *args, char *programname);

void reset_vpu(struct vpu *vpu);

// todo: test this
inline void vpu_kill_clones(struct vpu *vpu)
{
  unsigned n = vpu->child_count;

  vpu->exiting = 1;
  
  while(n--)
    if(vpu->child[n]->is_clone)
      delete_vpu(vpu->child[n]);

  vpu->exiting = 0;
  vpu->interrupt_wait = 0;
}

char *find_executable(char *program, unsigned bytecodeprog);

void vpu_syscall_exec(struct vpu *vpu)
{
  char *program = &vpu->data[vpu->data_segment][vpu->regs[1]];
  char *args = &vpu->data[vpu->data_segment][vpu->regs[2]];
  unsigned n, m;
  char programstr[82];
  FILE *fp;
  struct vpu *parent = vpu;
  char *found_program = 0;

  for(n=0;program[n] && memory_allowed(vpu, vpu->regs[1] + n);n++);
  for(m=0;args[m] && memory_allowed(vpu, vpu->regs[2] + m);m++);  

  if(!n || !memory_allowed(vpu, vpu->regs[1] + n) || !memory_allowed(vpu, vpu->regs[2] + m))
  {
    error:
    vpu->regs[0] = -1;
    return;
  }

  vpu_cwd(vpu);
  if(!strchr(program, '\\'))
  {
    found_program = find_executable(program, 1);
    if(!found_program)
      goto error;
    program = found_program;
  }

  if(!(fp = fopen(program, "rb")))
  {
    goto error;
  }
  fclose(fp);

  // if VPU is a clone, switch to parent vpu
  while(vpu->is_clone)
  {
    strcpy(vpu->parent->cwd, vpu->cwd);
    vpu = vpu->parent;
  }
  // then kill all clones
  vpu_kill_clones(vpu);

  free_vpu_args(vpu);
  
  vpu->argv = split_args(&vpu->argc, args, program);    
  strncpy(programstr, program, 81);
 
  
  while(vpu->dynsegment_count)
    free(vpu->data[vpu->dynsegments[--vpu->dynsegment_count].segment]);
  free(vpu->dynsegments);
  vpu->dynsegments=0;

  
  while(vpu->data_count)
    free(vpu->data[--vpu->data_count]);
  free(vpu->data);
  vpu->data=0;
  free(vpu->ports);
  vpu->ports = 0;  
  while(vpu->code_count)
    free(vpu->code[--vpu->code_count]);
  free(vpu->code);
  vpu->code=0;
  free(vpu->interrupts);
  vpu->interrupts=0;
  free(vpu->primary_stack);
  free(vpu->secondary_stack);
  vpu->primary_stack = vpu->secondary_stack = 0;

  if(load_program(programstr, vpu))
  {
    vpu->exiting = 1;
  }

  remove_fd_filepos_ptrs(vpu);

/*  memset(&vpu->regs, 0, VPU_REGISTER_COUNT*sizeof(unsigned short));
  memset(&vpu->flags, 0, sizeof(struct vpu_flags));
  vpu->compreg = 0;

  vpu->data_segment = 0;

  vpu->ip = 0;
  vpu->code_segment = 0;*/
  reset_vpu(vpu);

}

long timezone = 0;

void vpu_syscall_gettime(struct vpu *vpu)
{
  struct unixtime utime = unixtime(timezone);
//  vpu->regs[0] = utime.time>>48;
//  vpu->regs[1] = utime.time>>32;
  vpu->regs[0] = 0;
  vpu->regs[1] = 0;
  vpu->regs[2] = utime.time>>16;
  vpu->regs[3] = utime.time;

  vpu->regs[4] = utime.nanoseconds>>16;
  vpu->regs[5] = utime.nanoseconds;
}

void vpu_syscall_sound(struct vpu *vpu)
{
  sound(vpu->regs[1]);
}

void vpu_syscall_nosound(struct vpu *vpu)
{
  nosound();
}

void vpu_syscall_waitscrtrace(struct vpu *vpu)
{
  if(!vpu_has_active_window(vpu))
    return;

  if(videodriver == -1)
  {
    __asm
    {
      mov dx, 0x03da
      jump1:    
      in al, dx
      test al, 0x08
      jnz jump1
      jump2:
      in al, dx
      test al, 0x08
      jz jump2
    }
  }
  steps=0;
}

void vpu_syscall_sleep(struct vpu *vpu)
{
  vpu->pause = 1;
  vpu->sleep |= vpu->regs[4];
  vpu->sleep |= ((unsigned long)vpu->regs[3])<<16;
//  vpu->sleep |= vpu->regs[2]>>32;
//  vpu->sleep |= vpu->regs[1]>>48;
}

/*
  Returns pointer to the VPU struct by the given PID
*/

struct vpu *pid_to_vpu(unsigned pid)
{
  unsigned n = vpu_count;
  while(n--)
    if(vpus[n]->pid == pid)
      return vpus[n];
  return 0;
}


void vpu_syscall_sendsignal(struct vpu *vpu)
{
  struct vpu *target;
  if(!vpu->regs[1])
  {
    send_vpu_signal(vpu, vpu->regs[2]);
  }
  else if(target = pid_to_vpu(vpu->regs[1]))
  {
    if(!send_vpu_signal(target, vpu->regs[2]))
      vpu->regs[0] = 0;
      return;
  }
  vpu->regs[0] = 0;
}

void vpu_syscall_procinfo(struct vpu *vpu)
{
  char *process_name;
  unsigned n = window_count;
  static char zombie[] = "Zombie";
  
  if(vpu->regs[1] >= vpu_count)
  {
    vpu->regs[0] = 0;
    return;
  }
  else
  {
    vpu->regs[0] = vpus[vpu->regs[1]]->pid;
    if(vpus[vpu->regs[1]]->zombie)
      process_name = zombie;
    else if(vpus[vpu->regs[1]]->argc/* && *vpus[vpu->regs[1]]->argv[0]*/)
      process_name = vpus[vpu->regs[1]]->argv[0];
    else
      while(n--)
        if(windows[n]->function_pointer == vpuconsole_program
          && ((struct console_struct*)windows[n]->function_status)->vcpu == vpus[vpu->regs[1]])
        {
          process_name = windows[n]->title;
          break;
        }
    if(memory_allowed(vpu, vpu->regs[2]+strlen(process_name)+1))
      strcpy(&vpu->data[vpu->data_segment][vpu->regs[2]], process_name);                  
  }
}

extern char *cwd;
int chdrive(unsigned drive);

void vpu_syscall_chdir(struct vpu *vpu)
{
  unsigned n;
  char *dirname = &vpu->data[vpu->data_segment][vpu->regs[1]];
  for(n=0;dirname[n] && memory_allowed(vpu, vpu->regs[1] + n);n++);

  vpu_cwd(vpu);

  if(dirname[1] == ':')
    if(chdrive(dirname[0]))
      goto error;

  if(memory_allowed(vpu, vpu->regs[1] + n))
  {
    if(!chdir(dirname))
    {
//      strcpy(vpu->cwd, dirname);
      getcwd(&vpu->cwd, MAX_PATH);
      vpu->regs[0] = 0;
    }
    else
    {
      chdir(vpu->cwd);
      error:      
      vpu->regs[0] = 0xFFFF;
    }
  }
}

void vpu_syscall_opendir(struct vpu *vpu)
{
  unsigned dir_descriptor;
  unsigned n;
  char *dirname = &vpu->data[vpu->data_segment][vpu->regs[1]];
  char path[85];
  
  for(n=0;dirname[n] && memory_allowed(vpu, vpu->regs[1] + n);n++);

  if(!memory_allowed(vpu, vpu->regs[1] + n) || n>80)
    goto error;

  if(!*dirname)
    dirname = vpu->cwd;

  sprintf(path, "%s\\*.*", dirname);
  
  for(dir_descriptor=0;vpu->dir_handles[dir_descriptor] && dir_descriptor < MAX_DIRS;dir_descriptor++);

  if(dir_descriptor < MAX_DIRS)
  {
    vpu->dir_handles[dir_descriptor] = calloc(1, sizeof(struct dir_struct));
    if(!vpu->dir_handles[dir_descriptor])
    {
      nomemory();
      goto error;
    }
    if(_dos_findfirst(path, 0xFF, &vpu->dir_handles[dir_descriptor]->fileinfo))
    {
      free(vpu->dir_handles[dir_descriptor]);
      vpu->dir_handles[dir_descriptor] = 0;          
      goto error;
    }
    vpu->regs[0] = dir_descriptor;   
  }
  else
  {
    error:
    vpu->regs[0] = 0xFFFF;
  }
}

void vpu_syscall_readdir(struct vpu *vpu)
{
  if(vpu->regs[1] < MAX_DIRS && vpu->dir_handles[vpu->regs[1]] && !vpu->dir_handles[vpu->regs[1]]->error)
  {
    if(memory_allowed(vpu, vpu->regs[2] + strlen(vpu->dir_handles[vpu->regs[1]]->fileinfo.name)))
      strcpy(&vpu->data[vpu->data_segment][vpu->regs[2]], vpu->dir_handles[vpu->regs[1]]->fileinfo.name);
    vpu->regs[3] = vpu->dir_handles[vpu->regs[1]]->fileinfo.attrib;      
    vpu->regs[0] = 0;
    vpu->dir_handles[vpu->regs[1]]->error = _dos_findnext(&vpu->dir_handles[vpu->regs[1]]->fileinfo);
  }
  else
    vpu->regs[0] = 0xFFFF;
}

void vpu_syscall_closedir(struct vpu *vpu)
{
  if(vpu->regs[1] < MAX_DIRS)
  {
    free(vpu->dir_handles[vpu->regs[1]]);
    vpu->dir_handles[vpu->regs[1]] = 0;
  }
}

typedef struct { 

    unsigned short  twosecs : 5;    /* seconds / 2 */ 

    unsigned short  minutes : 6;    /* minutes (0,59) */ 

    unsigned short  hours   : 5;    /* hours (0,23) */ 

} ftime_t; 
typedef struct { 

    unsigned short  day     : 5;    /* day (1,31) */ 

    unsigned short  month   : 4;    /* month (1,12) */ 

    unsigned short  year    : 7;    /* 0 is 1980 */ 

} fdate_t; 
void vpu_syscall_stat(struct vpu *vpu)
{
  char *filename = &vpu->data[vpu->data_segment][vpu->regs[2]];
  unsigned char *return_struct = &vpu->data[vpu->data_segment][vpu->regs[1]];
  unsigned char return_offset = 0;
  unsigned short n;
  struct find_t fileinfo;
  unsigned year;
  unsigned char month, day, hour, min, sec;     

  for(n=0;filename[n] && memory_allowed(vpu, vpu->regs[2] + n);n++);
  if(!memory_allowed(vpu, vpu->regs[2] + n))
  {
    error:
    vpu->regs[0] = 0xFFFF;
    return;
  }

  if(_dos_findfirst(filename, 0xFF, &fileinfo))
    goto error;
  else
  {
    vpu->regs[0] = 0;
    if(memory_allowed(vpu, vpu->regs[1]+return_offset))
      *return_struct = fileinfo.attrib;
    return_offset++;
    if(memory_allowed(vpu, vpu->regs[1]+return_offset+sizeof(unsigned long)-1))
      *(unsigned long*)&return_struct[return_offset] = fileinfo.size;
    return_offset += sizeof(unsigned long);

    year = (*(fdate_t*)(&fileinfo.wr_date)).year+1980;
    month = (*(fdate_t*)(&fileinfo.wr_date)).month;
    day = (*(fdate_t*)(&fileinfo.wr_date)).day;
    hour = (*(ftime_t*)(&fileinfo.wr_time)).hours;
    min = (*(ftime_t*)(&fileinfo.wr_time)).minutes;
    sec = (*(ftime_t*)(&fileinfo.wr_time)).twosecs<<1;    
    
    if(memory_allowed(vpu, vpu->regs[1]+return_offset+sizeof(unsigned long long)-1))
      *(unsigned long long*)&return_struct[return_offset] = human2unixtime(year, month, day, hour, min, sec);
  }
}

void vpu_syscall_setpriority(struct vpu *vpu)
{
  if(!vpu->regs[1])
    vpu->priority = vpu->regs[2];
  else
    if(pid_to_vpu(vpu->regs[1]))
      pid_to_vpu(vpu->regs[1])->priority = vpu->regs[2];
}

void vpu_syscall_getpriority(struct vpu *vpu)
{
  if(!vpu->regs[1])
    vpu->regs[0] = vpu->priority;
  else
    if(pid_to_vpu(vpu->regs[1]))
      vpu->regs[0] = pid_to_vpu(vpu->regs[1])->priority;
}

void vpu_syscall_loadmod(struct vpu *vpu)
{
  unsigned n, m;
  char *filename = &vpu->data[vpu->data_segment][vpu->regs[1]];
  char *modname = &vpu->data[vpu->data_segment][vpu->regs[2]];

  for(n=0;filename[n] && memory_allowed(vpu, vpu->regs[1] + n);n++);
  for(m=0;modname[m] && memory_allowed(vpu, vpu->regs[2] + m);m++);
  if(!memory_allowed(vpu, vpu->regs[1] + n) || !memory_allowed(vpu, vpu->regs[2] + m))
  {
    vpu->regs[0] = 0xFFFF;
    return;
  }
  vpu->regs[0] = load_driver(filename, modname);  
}

void vpu_syscall_cmdmod(struct vpu *vpu)
{
  unsigned n, m;
  int i;
  char *devname = &vpu->data[vpu->data_segment][vpu->regs[1]];
  char *args = &vpu->data[vpu->data_segment][vpu->regs[2]];
  unsigned char devnum;

  for(n=0;devname[n] && memory_allowed(vpu, vpu->regs[1] + n);n++);
  for(m=0;args[m] && memory_allowed(vpu, vpu->regs[2] + m);m++);

  if(!memory_allowed(vpu, vpu->regs[1] + n) || !memory_allowed(vpu, vpu->regs[2] + m))
  {
    error:
    vpu->regs[0] = 0xFFFF;
    return;
  }

  if(driver_id(devname) < 0)
  {
    n = strlen(devname)-2;
    devnum = devname[n] - 0x30;
    devname[n] = 0;
  }

  if((i = driver_id(devname)) < 0 || (drivers[i].dev_count >= 0 && devnum > drivers[i].dev_count))
    goto error;

  if(!memory_allowed(vpu, vpu->regs[2] + drivers[i].cmd_func_arglen))
    goto error;

  dev_cmd(i, devnum, args);
}

void vpu_syscall_dup(struct vpu *vpu)
{
  unsigned oldfd = vpu->regs[1];
  unsigned file_descriptor;

  for(file_descriptor=0;vpu->file_handles[file_descriptor] && file_descriptor < MAX_FILES;file_descriptor++);

  if(file_descriptor < MAX_FILES && oldfd < MAX_FILES && vpu->file_handles[oldfd])
  {
    vpu_fd_makedup(vpu, file_descriptor, vpu, oldfd);
    
    vpu->regs[0] = file_descriptor;
  }
  else
    vpu->regs[0] = 0xFFFF;
}

void vpu_syscall_dup2(struct vpu *vpu)
{
  unsigned oldfd = vpu->regs[1];
  unsigned newfd = vpu->regs[2];

  if(oldfd < MAX_FILES && newfd < MAX_FILES && vpu->file_handles[oldfd])
  {
    if(oldfd == newfd)
    {
      vpu->regs[0] = newfd;
      return;
    }    
    if(vpu->file_handles[newfd])
      vpu_fd_close(vpu, newfd);

    vpu_fd_makedup(vpu, newfd, vpu, oldfd);                         
    
    vpu->regs[0] = newfd;
    
  }
  else
    vpu->regs[0] = 0xFFFF;  
}

void vpu_syscall_pipe(struct vpu *vpu)
{
  unsigned readend, writeend;

  for(readend=0;vpu->file_handles[readend] && readend < MAX_FILES;readend++);
  for(writeend=0;vpu->file_handles[writeend] && writeend < MAX_FILES || writeend == readend;writeend++);

  if(readend < MAX_FILES && writeend < MAX_FILES)
  {
    vpu->file_struct[readend]->buff = calloc(PIPE_BUF, sizeof(char));
    if(!vpu->file_struct[readend]->buff)
    {
      vpu->regs[0] = 0xFFFF;
      return;
    }     
    vpu->file_handles[readend] = (FILE*)-1;
    vpu->file_handles[writeend] = (FILE*)-1;
    vpu->file_struct[writeend]->pipe_to = vpu->file_struct[readend];
    vpu->file_struct[readend]->pipe_from = vpu->file_struct[writeend];
    vpu->file_struct[writeend]->is_ioport = 10;
    vpu->file_struct[readend]->is_ioport = 11;
//    vpu->file_struct[readend]->bufrpos = 0;
//    vpu->file_struct[writeend]->bufwpos = 0;
//    sprintf(message, "Avataan putki %p > %p puskuri %p", vpu->file_struct[writeend], vpu->file_struct[readend], vpu->file_struct[readend]->buff);
//    putstr(message);                                  
  }
  else
  {
    vpu->regs[0] = 0xFFFF;
    return;
  }
  vpu->regs[0] = 0;
  vpu->regs[1] = readend;
  vpu->regs[2] = writeend;
}

void vpu_syscall_mkdir(struct vpu *vpu)
{
  unsigned n;
  char *dirname = &vpu->data[vpu->data_segment][vpu->regs[1]];
  
  for(n=0;dirname[n] && memory_allowed(vpu, vpu->regs[1] + n);n++);

  if(!memory_allowed(vpu, vpu->regs[1] + n) || n>80)
  {
    vpu->regs[0] = 0xFFFF;
    return;
  }

  vpu->regs[0] = mkdir(dirname);
}

void vpu_syscall_sendfd(struct vpu *vpu)
{
  struct vpu *receiver = pid_to_vpu(vpu->regs[1]);
  unsigned oldfd = vpu->regs[2];
  unsigned newfd;

  if(!receiver)
  {
    error:
    vpu->regs[0] = 0xFFFF;
    steps = 0;
    return;
  }

  if(!vpu->file_handles[oldfd]) goto error;

  if(receiver->interrupt_count < SIGFDRECVD || !receiver->interrupts[SIGFDRECVD].address)
    goto error;

  for(newfd=0;receiver->file_handles[newfd];newfd++)
    if(newfd == MAX_FILES) goto error;

  if(send_vpu_signal(receiver, SIGFDRECVD))
    vpu_block(vpu);

  vpu_fd_makedup(receiver, newfd, vpu, vpu->regs[2]);

  receiver->regs[0] = newfd;

  vpu->regs[0] = 0;
}

void vpu_syscall_sethwint(struct vpu *vpu)
{
  unsigned hw_int = vpu->regs[1];
  unsigned vpu_int = vpu->regs[2];

  if(!set_vpu_hw_int(vpu, hw_int, vpu_int, vpu->regs[3]))
    vpu->regs[0] = 0xFFFF;
  else
    vpu->regs[0] = 0;
}

void vpu_syscall_unsethwint(struct vpu *vpu)
{
  unsigned hw_int = vpu->regs[1];
  unsigned n;

  for(n=vpu_hwint_count;n--;)
    if(vpu_hardware_int_table[n].vpu == vpu && vpu_hardware_int_table[n].hw_int == hw_int)
    {
      remove_vpu_hw_int(vpu, vpu_hardware_int_table[n].vpu_int);
      vpu->regs[0] = 0;
      return;
    }
  vpu->regs[0] = 0xFFFF;
}

void (*vpu_instr_syscall[])(struct vpu *) =
{
  vpu_syscall_read, // 0
  vpu_syscall_write, // 1
  vpu_syscall_open, // 2
  vpu_syscall_close, // 3
  vpu_syscall_chdir, // 4
  vpu_syscall_opendir, // 5
  vpu_syscall_readdir, // 6
  vpu_syscall_closedir, // 7
  vpu_syscall_lseek, // 8
  vpu_syscall_mmap, // 9
  vpu_syscall_stat, // 10
  vpu_syscall_munmap, // 11
  vpu_syscall_exit, // 12
  vpu_syscall_getargc, // 13
  vpu_syscall_getargv, // 14
  vpu_syscall_spawn, // 15
  vpu_syscall_clone, // 16
  vpu_syscall_fork, // 17
  vpu_syscall_gettime, // 18
  vpu_syscall_sound, // 19
  vpu_syscall_nosound, // 20
  vpu_syscall_waitscrtrace, // 21
  vpu_syscall_sleep, // 22
  vpu_syscall_sendsignal, // 23
  vpu_syscall_procinfo, // 24
  vpu_syscall_getpriority, // 25
  vpu_syscall_setpriority, // 26
  vpu_syscall_fsync, // 27
  vpu_syscall_loadmod, // 28
  vpu_syscall_cmdmod, // 29
  vpu_syscall_exec, // 30,
  vpu_syscall_dup, // 31,
  vpu_syscall_dup2, // 32,
  vpu_syscall_pipe, // 33
  vpu_syscall_mkdir, // 34
  vpu_syscall_sendfd, // 35
  vpu_syscall_sethwint, // 36
  vpu_syscall_unsethwint // 37  
};

void vpu_instr_sys(struct vpu *vpu)
{
  if(vpu->regs[0] < sizeof(vpu_instr_syscall) / sizeof(void*) && vpu_instr_syscall[vpu->regs[0]])
    (*vpu_instr_syscall[vpu->regs[0]])(vpu);
  else
    if(send_vpu_signal(vpu, SIGSYS))
      send_vpu_signal(vpu, SIGABRT);
}
