#define FIRST_FD 3
#include "vpustrct.c"

struct vpu **vpus;
unsigned vpu_count = 0;
unsigned pid_counter = 0;


int init_vpu(struct vpu *vpu, unsigned short stack_size, unsigned short code_size, unsigned short code_count, unsigned short 
data_size, unsigned short data_count, unsigned short interrupt_count)
{
  unsigned n;
  
  vpu->priority = DEFAULT_PRIORITY;
  vpu->sp = vpu->psp = vpu->ssp = stack_size-1;

  vpu->byteregs = (unsigned char*)vpu->regs;

  vpu->primary_stack = calloc(stack_size, sizeof(unsigned short));
  vpu->secondary_stack = calloc(stack_size, sizeof(unsigned short));
  if(!vpu->primary_stack || !vpu->secondary_stack)
    return 0;
  vpu->stack = vpu->primary_stack;

  vpu->stack_size = stack_size;    

  vpu->code = calloc(code_count, sizeof(char*));
  if(!vpu->code)
    return 0;

  vpu->realcode = vpu->code;

  while(vpu->code_count < code_count)
  {    
    vpu->code[vpu->code_count] = calloc(code_size+10, sizeof(char));
    if(!vpu->code[vpu->code_count])
      return 0;
    vpu->code_count++;
  }
  vpu->code_size = code_size;  

  vpu->data = calloc(data_count, sizeof(char*));
  if(!vpu->data)
    return 0;

  vpu->dataseg_table_len = data_count;  
  
  while(vpu->data_count < data_count)
  {
    vpu->data[vpu->data_count] = calloc(data_size, sizeof(char));
    if(!vpu->data[vpu->data_count])    
      return 0;
    vpu->data_count++;
  }

  vpu->data_size = data_size;  

  vpu->interrupts = calloc(interrupt_count, sizeof(struct vpu_interrupt));
  if(!vpu->interrupts)
    return 0;

  vpu->interrupt_count = interrupt_count;

  vpu->ports = calloc(VPU_PORT_COUNT, sizeof(unsigned short));
  if(!vpu->ports)
    return 0;
  vpu->port_count = VPU_PORT_COUNT;

  if(!vpu->file_handles)
  {
    vpu->file_handles = calloc(MAX_FILES, sizeof(FILE *));
    if(!vpu->file_handles) return 1;

    vpu->file_struct = calloc(MAX_FILES, sizeof(struct file_struct*));
    if(!vpu->file_struct) return 1;
    for(n=MAX_FILES;n--;)
    {
      vpu->file_struct[n] = calloc(1,sizeof(struct file_struct));
      if(!vpu->file_struct[n]) return 1;
    }

    vpu->dir_handles = calloc(MAX_DIRS, sizeof(struct dir_struct*));
    if(!vpu->dir_handles) return 1;
  }

  return 1;
}

struct vpu *spawn_new_vpu(void)
{
  void *new_pointer;
  if(!vpu_count++)
  {
    vpus = calloc(vpu_count, sizeof(struct vpu*));
    if(!vpus)
    {
      outofmemory0:
      vpu_count--;
      return 0;
    }
  }
  else
  {
    new_pointer = realloc(vpus, vpu_count*sizeof(struct vpu*));
    if(!new_pointer)
    {
      goto outofmemory0;
    }
    vpus = new_pointer;    
  }
  if(!(vpus[vpu_count-1] = calloc(1, sizeof(struct vpu))))
  {
    if(--vpu_count)
      vpus = realloc(vpus, vpu_count*sizeof(struct vpu*));
    else
      free(vpus);      
    return 0;
  }
  vpus[vpu_count-1]->pid = ++pid_counter;
  return vpus[vpu_count-1];  
}

void delete_child(struct vpu *child)
{
  struct vpu *parent = child->parent;
  unsigned n=parent->child_count--;

  if(!parent->child_count)
  {
    free(parent->child);
    parent->child=0;
    return;
  }
  while(n--)
    if(parent->child[n] == child)
      break;
  for(;n<parent->child_count;n++)
    parent->child[n] = parent->child[n+1];

  parent->child = realloc(parent->child, parent->child_count* sizeof(struct vpu*));
}

inline int vpu_shares_args(struct vpu *vpu)
{
  unsigned n = vpu_count;
  while(n--)
    if(vpus[n] != vpu && vpu->argv == vpus[n]->argv)
      return 1;
  return 0;
}

void free_vpu_args(struct vpu *vpu)
{  
  if(vpu_shares_args(vpu))
  {
    vpu->argc = 0;
    return;
  }
      
  while(vpu->argc)
    free(vpu->argv[--vpu->argc]);
  free(vpu->argv);
  vpu->argv=0;
}

void free_vpu_resources(struct vpu *vpu)
{
  unsigned n;

  for(n=vpu_hwint_count;n--;)
    if(vpu_hardware_int_table[n].vpu == vpu)
      remove_vpu_hw_int(vpu, vpu_hardware_int_table[n].vpu_int);  
  
  if(vpu->is_clone) goto clonekill;

  free_vpu_args(vpu);

  if(vpu->file_handles)
    for(n = MAX_FILES;n--;)
      if(vpu->file_handles[n])
        vpu_fd_close(vpu, n);
  free(vpu->file_handles);
   
  if(vpu->dir_handles)        
    for(n = MAX_DIRS;n--;)
      free(vpu->dir_handles[n]);
  free(vpu->dir_handles);

  if(vpu->file_struct)  
    for(n=MAX_FILES;n--;)
      free(vpu->file_struct[n]);
  free(vpu->file_struct);

  while(vpu->dynsegment_count--)
    free(vpu->data[vpu->dynsegments[vpu->dynsegment_count].segment]);
  free(vpu->dynsegments);
 
  while(vpu->data_count--)
    free(vpu->data[vpu->data_count]);
  free(vpu->data);

  while(vpu->code_count--)
    free(vpu->code[vpu->code_count]);
  free(vpu->code);

  free(vpu->interrupts);
  free(vpu->ports);

  clonekill:
  free(vpu->primary_stack);  
  free(vpu->secondary_stack);

  vpu->zombie = 1;
}

int send_termination_signal(struct vpu *vpu)
{
  if(vpu->parent)
  {
    if(vpu->parent->interrupt_wait && !vpu->parent->exiting)
    {
//      putstr("parent already processing signal");
      return 1;
    }
    else
    {
      if(vpu->termination_signal)
        vpu->parent->interrupt_wait = vpu->termination_signal;
      vpu->termination_signal = 0;
//      putstr("Sent termination signal");
    }
  }
  return 0;
}

void delete_vpu(struct vpu *vpu)
{
  unsigned n;

  if(!vpu) return;

  if(!vpu->zombie)
    free_vpu_resources(vpu);

  vpu->exiting=1;

  while(vpu->child_count)
    delete_vpu(vpu->child[0]);
 
  if(vpu->parent)
  {
    if(send_termination_signal(vpu))
      return;
//    putstr("deleting child...");
    delete_child(vpu);
  }

  free_vpu_args(vpu);

  n=vpu_count;
  while(vpus[--n] != vpu);
  while(n<vpu_count-1)
  {
    vpus[n] = vpus[n+1];
    n++; 
  }

  free(vpu);
   
  if(--vpu_count)
    vpus = realloc(vpus, vpu_count*sizeof(struct vpu*));
  else
  {
    free(vpus);
    vpus = 0;
  }
//  putstr("vpu deleted");
}

