//#include "../globals.c"

int vpuconsole_program(struct prog_window *window, int args, int *keypress);
struct prog_window *init_vpuprog_window(unsigned x, unsigned y, unsigned char *title, unsigned mouse_int, unsigned mouse_port, void *vpu);
struct window_element *window_text(struct prog_window *window, unsigned int x, unsigned int y, unsigned char *text, unsigned char color);
struct window_element *window_button(struct prog_window *window, unsigned int x, unsigned int y, unsigned int width, unsigned int height, unsigned char *text, unsigned char color);
struct window_element *window_textbox(struct prog_window *window, unsigned int x, unsigned int y, unsigned int width, unsigned char *text, unsigned int len);
struct window_element *window_list(struct prog_window *window, unsigned int x, unsigned int y, unsigned int width, unsigned int height, unsigned char **items, unsigned int len);
struct window_element *window_image(struct prog_window *window, unsigned int x, unsigned int y, unsigned int width, unsigned int height, char *filename);
struct window_element *window_scrollbar(struct prog_window *window, unsigned int x, unsigned int y, unsigned type, unsigned int length, unsigned int count, int element, int *value);
struct window_element *window_icon(struct prog_window *window, unsigned int x, unsigned int y, unsigned char *text, char *icon_file);
int vpuwindow_program(struct prog_window *window, int args, int *keypress);
void put_pixel(unsigned int x, unsigned int y, unsigned char color);
extern struct prog_window *running_window;
void draw_line(unsigned int x0, unsigned int y0, unsigned int x1, unsigned int y1, unsigned char color);
void draw_rectangle(int x0, int y0, unsigned int width, unsigned int height, unsigned char color);
void draw_circle(unsigned x, unsigned y, unsigned r, unsigned char color);
extern unsigned char _FONT_HEIGHT;
extern unsigned char _WINDOW_BGCOLOR;
void gprint_text(unsigned int x, unsigned int y, unsigned char __far *text, unsigned char color, unsigned char font_height, unsigned int x2, unsigned monospace);
void draw_window(struct prog_window *window, unsigned active, unsigned update);
void draw_mouse_cursor(unsigned updatebg);
int *get_mouse_movement(void);
unsigned char windowpos_visible(struct prog_window *window, int x, int y);
void draw_screen(void);
void video_mode(unsigned char mode);
int dialog(unsigned int x, unsigned int y, unsigned int width, unsigned int height, unsigned type, unsigned help_button, unsigned sound, unsigned char *message, unsigned char *title, unsigned char *helptext);
extern struct prog_window **windows;
extern int active_window;
extern unsigned int _RES_Y;
extern unsigned int _RES_X;
extern volatile unsigned char video;
extern unsigned char __far *VGA;
extern unsigned window_count;
void set_active_window(int window, unsigned draw);
extern int *window_visibility_order;

char nowindowerror[] = "No such window!";

void vpu_scrcall_createwindow(struct vpu *vpu)
{
  char *title = &vpu->data[vpu->data_segment][vpu->regs[3]];
  unsigned short window_descriptor;


  if(!memory_allowed(vpu, vpu->regs[3]+strlen(title)))
    return;

  for(window_descriptor=0;vpu->windows[window_descriptor].window && window_descriptor < MAX_WINDOWS;window_descriptor++);

  if(window_descriptor < MAX_WINDOWS-1)
  {
    vpu->windows[window_descriptor].window = init_vpuprog_window(vpu->regs[1], vpu->regs[2], title, vpu->regs[6], vpu->regs[7], vpu);
    if(!vpu->windows[window_descriptor].window)
      goto error;
    vpu->regs[0] = window_descriptor;
    vpu->windows[window_descriptor].keypress_int = vpu->regs[4];
    vpu->windows[window_descriptor].keypress_port = vpu->regs[5];
  }
  else
    error:vpu->regs[0] = 0xFFFF;
}

void vpu_scrcall_fullscreen(struct vpu *vpu)
{
  const unsigned char allow_modes[] = { 0x06, 0x0F, 0x10, 0x11, 0x12, 0x13, 0x00 };
//  unsigned long buffer_size = 0;
  vpu_scrcall_createwindow(vpu);
  vpu->windows[vpu->regs[0]].window->fullscreen=1;
  if(vpu->regs[1] && strchr(allow_modes, vpu->regs[1]))  
    vpu->windows[vpu->regs[0]].window->scrmode = vpu->regs[1];
  else
    vpu->windows[vpu->regs[0]].window->scrmode = video;
  video_mode(vpu->windows[vpu->regs[0]].window->scrmode);
  vpu->windows[vpu->regs[0]].window->width = _RES_X;
  vpu->windows[vpu->regs[0]].window->height = _RES_Y;
  vpu->windows[vpu->regs[0]].window->x = 0;
  vpu->windows[vpu->regs[0]].window->y = 0;
  switch(vpu->regs[1])
  {
    case 0x13:
      vpu->windows[vpu->regs[0]].buffer_size = 64000;
      break;
    case 0x06:
      vpu->windows[vpu->regs[0]].buffer_size = 16384;
      break;
    case 0x0F:
      vpu->windows[vpu->regs[0]].buffer_size = 28000;
      break;
    case 0x11:
      vpu->windows[vpu->regs[0]].buffer_size = 38400;
  }
  if(vpu->windows[vpu->regs[0]].buffer_size)
    vpu->windows[vpu->regs[0]].renderer = _fcalloc(vpu->windows[vpu->regs[0]].buffer_size, sizeof(unsigned char));  
  draw_screen();
}

void vpu_scrcall_putpixel(struct vpu *vpu)
{
  unsigned x, y; 

  x = vpu->windows[vpu->regs[4]].window->x + vpu->regs[1];
  y = vpu->windows[vpu->regs[4]].window->y + vpu->regs[2] + _FONT_HEIGHT+2;

//  if(windowpos_visible(vpu->windows[vpu->regs[4]].window, x, y))
  put_pixel(x, y, vpu->regs[3]);
}

void vpu_scrcall_memset(struct vpu *vpu)
{
  unsigned color = vpu->regs[3];
  unsigned count = vpu->regs[5];
  unsigned maxx = vpu->windows[vpu->regs[4]].window->x + vpu->windows[vpu->regs[4]].window->width;
  unsigned maxy = vpu->windows[vpu->regs[4]].window->y + vpu->windows[vpu->regs[4]].window->height+_FONT_HEIGHT+2;
  unsigned x = vpu->windows[vpu->regs[4]].window->x+vpu->regs[1];
  unsigned y = vpu->windows[vpu->regs[4]].window->y+vpu->regs[2]+_FONT_HEIGHT+2;

  while(count--)
  {
    if(x >= maxx)
    {
      x = vpu->windows[vpu->regs[4]].window->x;
      y++;
    }
    if(y >= maxy)
      return;
    put_pixel(x++, y, color);
  }
}

void vpu_scrcall_memcpy(struct vpu *vpu)
{
  unsigned count = vpu->regs[5];
  unsigned dataoffset = vpu->regs[3];
  unsigned char *dataptr = &vpu->data[vpu->data_segment][dataoffset];
  unsigned maxx = vpu->windows[vpu->regs[4]].window->x + vpu->windows[vpu->regs[4]].window->width;
  unsigned maxy = vpu->windows[vpu->regs[4]].window->y + vpu->windows[vpu->regs[4]].window->height+_FONT_HEIGHT+2;
  unsigned x = vpu->windows[vpu->regs[4]].window->x+vpu->regs[1];
  unsigned y = vpu->windows[vpu->regs[4]].window->y+vpu->regs[2]+_FONT_HEIGHT+2;

  if(!memory_allowed(vpu, dataoffset) || !memory_allowed(vpu, dataoffset+count))
    return;

  while(count--)
  {
    if(x >= maxx)
    {
      x = vpu->windows[vpu->regs[4]].window->x;
      y++;
    }
    if(y >= maxy)
      return;
    put_pixel(x++, y, *dataptr++);
  }  
}

void vpu_scrcall_update(struct vpu *vpu)
{
  if(!vpu->windows[vpu->regs[4]].window)
    putstr(nowindowerror);
  else if(!vpu->windows[vpu->regs[4]].window->minimized && vpu->windows[vpu->regs[4]].renderer)
  {
    draw_mouse_cursor(2);      
    memcpy(VGA, vpu->windows[vpu->regs[4]].renderer, vpu->windows[vpu->regs[4]].buffer_size);
    draw_mouse_cursor(0);
    steps = 0;
  }
}

void vpu_scrcall_clear(struct vpu *vpu)
{
//  draw_mouse_cursor(2);    
  if(!vpu->windows[vpu->regs[4]].window->minimized
    && windows[active_window] == vpu->windows[vpu->regs[4]].window)
  {
    draw_mouse_cursor(2);        
    draw_window(vpu->windows[vpu->regs[4]].window, 0, 1);
    draw_mouse_cursor(0);
  }
}

void vpu_scrcall_text(struct vpu *vpu)
{
  char *text = &vpu->data[vpu->data_segment][vpu->regs[3]];
  struct window_element *element;

  if(!memory_allowed(vpu, vpu->regs[3]+strlen(text)))
    return;

  element = window_text(vpu->windows[vpu->regs[4]].window, vpu->regs[1], vpu->regs[2], text, vpu->regs[5]);
  if(!element) vpu->regs[0] = 0xFFFF;
  draw_window(vpu->windows[vpu->regs[4]].window, 0, 1);    
}

void vpu_scrcall_button(struct vpu *vpu)
{
  char *text = &vpu->data[vpu->data_segment][vpu->regs[3]];
  struct window_element *element;

  if(!memory_allowed(vpu, vpu->regs[3]+strlen(text)))
    return;

  element = window_button(vpu->windows[vpu->regs[4]].window, vpu->regs[1], vpu->regs[2], vpu->regs[6], vpu->regs[7],
    text, vpu->regs[5]);
  
  if(!element) vpu->regs[0] = 0xFFFF;
  else vpu->regs[0] = 0x00FF + vpu->windows[vpu->regs[4]].window->element_count;
  
  draw_window(vpu->windows[vpu->regs[4]].window, 0, 1);    
}

void vpu_scrcall_textbox_update(struct vpu *vpu)
{
  char *text = &vpu->data[vpu->data_segment][vpu->regs[3]];
  struct window_element *element;  
  if(!memory_allowed(vpu, vpu->regs[3]+strlen(text)) || vpu->regs[1] >= vpu->windows[vpu->regs[4]].window->element_count)
    return;

  element = vpu->windows[vpu->regs[4]].window->window_grid[vpu->regs[1]];
  free(element->text);
  element->text = calloc(element->c+1, sizeof(char)); // this calloc should not be able to fail...
  strncpy(element->text, text, element->c);
}

void vpu_scrcall_textbox(struct vpu *vpu)
{
  char *text = &vpu->data[vpu->data_segment][vpu->regs[3]];
  struct window_element *element;

  if(!memory_allowed(vpu, vpu->regs[3]+vpu->regs[5]))
    return;

  element = window_textbox(vpu->windows[vpu->regs[4]].window, vpu->regs[1], vpu->regs[2], vpu->regs[6], text,
    vpu->regs[5]);
  
  if(!element) vpu->regs[0] = 0xFFFF;
  else
  {
    vpu->regs[0] = 0x00FF + vpu->windows[vpu->regs[4]].window->element_count;
    vpu->regs[1] = vpu->windows[vpu->regs[4]].window->element_count-1;
  }
  

  element->connect_value = text;
  
  draw_window(vpu->windows[vpu->regs[4]].window, 0, 1);    
}

unsigned char **make_list(struct vpu *vpu, unsigned list_len, unsigned short *items)
{
  unsigned char **retval = calloc(list_len, sizeof(char *));
  if(!retval)
  {
    nomemory:
    vpu->regs[0] = 0xFFFF;
    return 0;
  }

  while(list_len--)
    if(memory_allowed(vpu, items[list_len]))
    {
      
      if(!(retval[list_len] = calloc(1+strlen(&vpu->data[vpu->data_segment][items[list_len]]), sizeof(unsigned char))))
      {
        while(++list_len < vpu->regs[7])
          free(retval[list_len]);
        free(retval);
        goto nomemory;
      }
      strcpy(retval[list_len], &vpu->data[vpu->data_segment][items[list_len]]);
    }
  return retval;
}

void vpu_scrcall_list_textbox(struct vpu *vpu, unsigned char type)
{
  struct window_element *element;
  unsigned short *items = (unsigned short*)&vpu->data[vpu->data_segment][vpu->regs[3]];
  unsigned short list_len = vpu->regs[7];
  unsigned char **list_rows;
  

  if(!memory_allowed(vpu, vpu->regs[3]+(list_len<<1)))
    return;

  element = window_list(vpu->windows[vpu->regs[4]].window, vpu->regs[1], vpu->regs[2], vpu->regs[5], vpu->regs[6], 0, vpu->regs[7]);
  if(!element)
  {
    nomemory:
    vpu->regs[0] = 0xFFFF;
    return;
  }
  
  list_rows = make_list(vpu, list_len, items);
  if(list_rows)
    element->items = list_rows;
  else
    goto nomemory;

//  free(list_rows);

  if(type)
    element->active = 0;

  vpu->regs[0] = 0x00FF + vpu->windows[vpu->regs[4]].window->element_count;
  vpu->regs[1] = vpu->windows[vpu->regs[4]].window->element_count-1;
}

void vpu_scrcall_list(struct vpu *vpu)
{
  vpu_scrcall_list_textbox(vpu, 0);
}

void vpu_scrcall_mltext(struct vpu *vpu)
{
  vpu_scrcall_list_textbox(vpu, 1);
}

void vpu_scrcall_updatemltext(struct vpu *vpu)
{
  struct window_element *element;
  unsigned short *items = (unsigned short*)&vpu->data[vpu->data_segment][vpu->regs[3]];
  unsigned short list_len = vpu->regs[7];
  unsigned char **list_rows;

  if(!memory_allowed(vpu, vpu->regs[3]+(list_len<<1)) || vpu->regs[1] >= vpu->windows[vpu->regs[4]].window->element_count)
    return;  
    
  element = vpu->windows[vpu->regs[4]].window->window_grid[vpu->regs[1]];

  while(element->c--)
    free(element->items[element->c]);
  free(element->items);

  list_rows = make_list(vpu, list_len, items);
  if(list_rows)
  {
    element->items = list_rows;
    element->c = vpu->regs[7];
  }
  else
    vpu->regs[0] = 0xFFFF;  
}

void vpu_scrcall_scrollbar(struct vpu *vpu)
{
  struct prog_window *window = vpu->windows[vpu->regs[4]].window;
  unsigned x = vpu->regs[1];
  unsigned y = vpu->regs[2];
  // types: 0 = vertical scrollbar, bind value to a list's vertical scroll
  // 1: horizontal scrollbar, bind value to a list's horizontal scroll
  // 2: vertical scrollbar, read or write value from a port in element reg
  // 3: horizontal scrollbar, read or write the value from a port
  unsigned type = vpu->regs[3];
  unsigned length = vpu->regs[5];
  unsigned count = vpu->regs[6];
  unsigned read_value = vpu->regs[7];
  struct window_element *element;
  int *connect_value;
  struct window_element *return_element;
  

  if(type < 2)
  {
    if(read_value >= window->element_count)
    {
      error:vpu->regs[0] = -1;
      return;
    }
    element = window->window_grid[vpu->regs[7]];
    if(!type)
      connect_value = (int*)&window->window_grid[read_value]->scroll_y;
    else
      connect_value = (int*)&window->window_grid[read_value]->scroll_x;
  }
  else
  {
    element = 0;
    if(read_value >= VPU_PORT_COUNT) goto error;
    connect_value = (int*)&vpu->ports[read_value];
  }

  return_element = window_scrollbar(window, x, y, type, length, count, read_value, connect_value);
  if(!return_element) goto error;
  vpu->regs[0] = 0x00FF + window->element_count;
  draw_window(window, 1, 3);
}

void vpu_scrcall_icon(struct vpu *vpu)
{
  char *text = &vpu->data[vpu->data_segment][vpu->regs[3]];
  char *icon = &vpu->data[vpu->data_segment][vpu->regs[5]];
  struct window_element *element;

  if(!memory_allowed(vpu, vpu->regs[3]+strlen(text)) || !memory_allowed(vpu, vpu->regs[5]+strlen(icon)))
    return;

  element = window_icon(vpu->windows[vpu->regs[4]].window, vpu->regs[1], vpu->regs[2], text, icon);
  
  if(!element) vpu->regs[0] = 0xFFFF;
  else vpu->regs[0] = 0x00FF + vpu->windows[vpu->regs[4]].window->element_count;
  
  draw_window(vpu->windows[vpu->regs[4]].window, 0, 1);    
}

void vpu_scrcall_circle(struct vpu *vpu)
{
  unsigned x, y, r; 

  x = vpu->windows[vpu->regs[4]].window->x + vpu->regs[1];
  y = vpu->windows[vpu->regs[4]].window->y + vpu->regs[2] + _FONT_HEIGHT+2;
  r = vpu->regs[5];

  draw_circle(x, y, r, vpu->regs[3]);
}

void vpu_scrcall_line(struct vpu *vpu)
{
  unsigned x0, y0, x1, y1;
  x0 = vpu->windows[vpu->regs[4]].window->x + vpu->regs[1];
  y0 = vpu->windows[vpu->regs[4]].window->y + vpu->regs[2] + _FONT_HEIGHT+2;
  x1 = vpu->windows[vpu->regs[4]].window->x + vpu->regs[5];
  y1 = vpu->windows[vpu->regs[4]].window->y + vpu->regs[6] + _FONT_HEIGHT+2;

  draw_line(x0, y0, x1, y1, vpu->regs[3]); 
}

void vpu_scrcall_rectangle(struct vpu *vpu)
{
  unsigned x, y, w, h;

  x = vpu->windows[vpu->regs[4]].window->x + vpu->regs[1];
  y = vpu->windows[vpu->regs[4]].window->y + vpu->regs[2] + _FONT_HEIGHT+2;
  w = vpu->regs[5];
  h = vpu->regs[6];

  draw_rectangle(x, y, w, h, vpu->regs[3]);
}

void vpu_scrcall_temptext(struct vpu *vpu)
{
  char *text = &vpu->data[vpu->data_segment][vpu->regs[3]];
  unsigned x = vpu->windows[vpu->regs[4]].window->x + vpu->regs[1];
  unsigned y = vpu->windows[vpu->regs[4]].window->y + vpu->regs[2] + _FONT_HEIGHT+2;
  unsigned x2 = vpu->windows[vpu->regs[4]].window->x+vpu->windows[vpu->regs[4]].window->width;

  if(!memory_allowed(vpu, vpu->regs[3]+strlen(text)))
    return;

  gprint_text(x, y, text, vpu->regs[5], _FONT_HEIGHT, x2, vpu->regs[6]);
}

/*
  BX == title
  CX == message
  DX == help
  EL == type
  EH == sound
*/
void vpu_scrcall_dialog(struct vpu *vpu)
{
  char *title = &vpu->data[vpu->data_segment][vpu->regs[1]];
  char *message = &vpu->data[vpu->data_segment][vpu->regs[2]];
  char *helptext = 0;

  if(!memory_allowed(vpu, vpu->regs[2]+strlen(message)) && !memory_allowed(vpu, vpu->regs[1]+strlen(title)))
    return;
  if(vpu->regs[3] && memory_allowed(vpu, vpu->regs[3]+strlen(helptext)))
    helptext = &vpu->data[vpu->data_segment][vpu->regs[3]];

  
  vpu->regs[0] = dialog(-1, -1, 170+(strlen(message)<<3), 50, vpu->regs[4]&0x00FF, vpu->regs[3], (vpu->regs[4]&0xFF00)>>8, message, title, helptext);
//  vpu->regs[0] = dialog(-1, -1, 170+(strlen(message)<<3), 50, vpu->regs[4]&0x00FF, vpu->regs[3], 1, message, title, helptext);
}

void vpu_scrcall_xpm2img(struct vpu *vpu)
{
  char *filename = &vpu->data[vpu->data_segment][vpu->regs[3]];
  struct window_element *element;

  if(!memory_allowed(vpu, vpu->regs[3]+strlen(filename)))
    return;

  element = window_image(vpu->windows[vpu->regs[4]].window, vpu->regs[1], vpu->regs[2], vpu->regs[5], vpu->regs[6], filename);
  if(!element) vpu->regs[0] = 0xFFFF;
  draw_window(vpu->windows[vpu->regs[4]].window, 0, 1);      
}

void (*vpu_instr_scrcall[])(struct vpu *) =
{                            // AX ==
  vpu_scrcall_createwindow, // 0
  vpu_scrcall_fullscreen, // 1
  vpu_scrcall_update, // 2
  vpu_scrcall_clear, // 3
  vpu_scrcall_text, // 4
  vpu_scrcall_putpixel, // 5
  vpu_scrcall_memset, // 6
  vpu_scrcall_memcpy, // 7
  vpu_scrcall_button, // 8
  vpu_scrcall_textbox, // 9
  vpu_scrcall_list, // 10
  vpu_scrcall_scrollbar, // 11
  vpu_scrcall_icon, // 12
  vpu_scrcall_circle, // 13
  vpu_scrcall_line, // 14
  vpu_scrcall_rectangle, // 15
  vpu_scrcall_mltext, // 16
  vpu_scrcall_updatemltext, // 17
  vpu_scrcall_temptext, // 18
  vpu_scrcall_textbox_update, // 19
  vpu_scrcall_dialog, // 20
  vpu_scrcall_xpm2img // 21
};

void vpu_instr_scr(struct vpu *vpu)
{
  struct prog_window *oldrw = running_window;
  unsigned char __far *vgap;
  if(vpu->regs[0] < sizeof(vpu_instr_scrcall) / sizeof(void*))
  {
    if(vpu->regs[0] > 2 && vpu_instr_scrcall[vpu->regs[0]] != vpu_scrcall_dialog)
    {
      if(vpu->windows[vpu->regs[4]].window)
      {
        vgap = VGA;
        running_window = vpu->windows[vpu->regs[4]].window;      
        if(vpu->windows[vpu->regs[4]].renderer && !vpu->windows[vpu->regs[4]].window->minimized)
          VGA = vpu->windows[vpu->regs[4]].renderer;
        (*vpu_instr_scrcall[vpu->regs[0]])(vpu);
        running_window = oldrw;
        VGA = vgap;      
      }
      else
        putstr(nowindowerror);      
    }
    else
      (*vpu_instr_scrcall[vpu->regs[0]])(vpu);    
  }
}

void vpu_getwindowkeys(struct vpu *vpu)
{
  unsigned n = MAX_WINDOWS;
  static unsigned char specialkey = 0;
  {
    while(n--)
      if(vpu->windows[n].window && vpu->windows[n].keypress_int && !vpu->windows[n].window->minimized)
      {
        if(specialkey && vpu->ports[vpu->windows[n].keypress_port] == -1)
        {
          vpu->ports[vpu->windows[n].keypress_port] = ((struct vpu_progwindow *)vpu->windows[n].window->function_status)->keybuff[1];
          specialkey = 0;
        }
        else if(( (struct vpu_progwindow *)vpu->windows[n].window->function_status )->keybuff[0] != -1)
        {
          vpu->interrupt_wait = vpu->windows[n].keypress_int;
          vpu->ports[vpu->windows[n].keypress_port] = ((struct vpu_progwindow*)vpu->windows[n].window->function_status)->keybuff[0];
          if(!((struct vpu_progwindow *)vpu->windows[n].window->function_status)->keybuff[0])
            specialkey = 1;             
          ((struct vpu_progwindow *)vpu->windows[n].window->function_status)->keybuff[0] = -1;
        }
        break;
      }
  }
}
